/*____________________________________________________________________________*\ *
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/HTTP/DirIndex.cpp,v $
 * $Date: 2003/05/31 10:34:24 $
 *
 Description:
	Directory Handler. Flexible generation of directory indexes.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <ctype.h>
#include <stdio.h>
#include <fstream.h>

#include "DeQuote.h"
#include "PIStrStr.h"
#include "StrToken.h"
#include "SwapStr.h"

#include "Pi3Expr.h"
#include "HandBase.h"
#include "HTTPCore.h"
#include "HTTPUtil.h"

/*____________________________________________________________________________*\
 *
 Description:
	Documentation.
\*____________________________________________________________________________*/
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	DirectoryIndex

Description:
	Generate directory indexes based on flexible criteria and send the
	output directly to the remote client. HTML directory indexes are
	optional sortable and can tag files, which are newer than a threshold
	date.

	The directory listing sent to the user is composed of the following
	document in order
<LIST>
<ITEM>An optional header, specified by 'HeaderPattern'.
<ITEM>An optional header file from the directory, specified by 'HeaderFile'.
<ITEM>An optional list start pattern, specified by 'ListTop'.
<ITEM>Every applicable file in the directory, output format is specified
by 'FilePattern'.
<ITEM>An optional list end pattern, specified by 'ListBottom'.
<ITEM>An optional footer file, specified by 'FooterFile'.
<ITEM>An optional footer pattern, specified by 'FooterPattern'.
</LIST>

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>HeaderPattern
<TD>-
<TD>A Pi3Expression
<TD>Output first
<TD>HeaderPattern="<HTML><TITLE>Index of %p</TITLE>"

<TR>
<TD>HeaderFile
<TD>-
<TD>A filename
<TD>File sent before listing
<TD>HeaderFile="HEADER"

<TR>
<TR>
<TD>FooterFile
<TD>-
<TD>A filename
<TD>File send after listing
<TD>FooterFile="FOOTER"

<TR>
<TD>ListTop
<TD>-
<TD>A Pi3Expression
<TD>Output before file iteration
<TD>ListTop="<H2>Directory listing of %p</H2>"

<TR>
<TD>FilePattern
<TD>+
<TD>A Pi3Expression
<TD>Determines output generated for every file
<TD>FilePattern="&abbrev(%n,32,...)&align(%s,10,right)"

<TR>
<TD>ListBottom
<TD>-
<TD>A Pi3Expression
<TD>Output after file iteration
<TD>ListBottom="<HR>"

<TR>
<TD>FooterPattern
<TD>-
<TD>A Pi3Expression
<TD>Output last
<TD>FooterPattern="</HTML>"

<TR>
<TD>DescriptionFile
<TD>-
<TD>&lt;filename&gt;
<TD>File with descriptions of files
<TD>DescriptionFile=".desc"

<TR>
<TD>MIMEIcon
<TD>-
<TD>"MIME icon_relative_path"
<TD>Maps a MIME specification to an icons path
<TD>MIMEIcon="text/* /icons/text.gif"; MIMEIcon="* /icons/unknown.gif"

<TR>
<TD>Include
<TD>-
<TD>Regular expression
<TD>A file pattern to include
<TD>Include="*"

<TR>
<TD>Exclude
<TD>-
<TD>Regular expression
<TD>A file pattern to exclude
<TD>Exclude="."

<TR>
<TD>HTTPEquiv
<TD>-
<TD>&lt;rfc822-variable&gt;: &lt;content&gt;
<TD>An rfc822 response header to send
<TD>HTTPEquiv="Content-Type: text/plain"

<TR>
<TD>LastModifiedFormat
<TD>-
<TD>"%a %b %d %Y %T GMT"
<TD>Format for file last modified date
<TD>LastModifedFormat="%a %b %d %Y %T GMT"

<TR>
<TD>SwapFileName
<TD>-
<TD>"file1 file2"
<TD>Replace any file1 name with file2
<TD>SwapFileName=".. Parent Directory"

<TR>
<TD>NewPattern
<TD>-
<TD>A wellformed HTML fragment
<TD>Replaces the %m parameter
<TD>NewPattern "&lt;IMG SRC="/icons/new.gif"&gt;"

<TR>
<TD>NewThreshold
<TD>-
<TD><number>
<TD>Threshold for new files in days
<TD>NewThreshold 365

<TR>
<TD>Options
<TD>-
<TD>Options flags as described below
<TD>Options which effect behaviour
<TD>Options "Name | AltName | Size etc."

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>

<H5>
	HeaderPattern
</H5>
Specifies the pattern that will be output first. See the directive
'FilePattern' for a description of parameters.

<H5>
	FooterPattern
</H5>
Specifies the pattern that will be output last. See the directive
'FilePattern' for a description of parameters.

<H5>
	HeaderFile
</H5>
Specifies the name of a file that will be sent before the directory listing
and ListTop pattern but after the HeaderPattern pattern.
Typically this file would also be excluded
from the directory list by naming it in an 'Exclude' directive.

<H5>
	FooterFile
</H5>
Specified the name of a file that will be sent after the directory
index and ListBottom pattern, but before the FooterPattern pattern.
Typically this file would also be excluded
from the directory listing by naming it in an 'Exclude' directive.

<H5>
	ListTop
</H5>
Determines the output generated before iterating through files. See 
'FilePattern' for a description of parameters.

<H5>
	FilePattern
</H5>
Determines the output generated for every file which meets the criteria
specified by 'Include' and 'Exclude'. This is a Pi3Expression with
the following parameters specific to directory indexing:
<CENTER>
<TABLE BORDER=1>
<TH>Pi3Expression parameter
<TH>Evaluates to

<TR><TD>%a<TD>Alternate file name
<TR><TD>%b<TD>Empty if file is not a directory
<TR><TD>%c<TD>Media (MIME) type
<TR><TD>%d<TD>Description of the file if available
<TR><TD>%l<TD>File last modified date
<TR><TD>%m<TD>Mark file conditional as new
<TR><TD>%n<TD>File/directory name
<TR><TD>%i<TD>Icon
<TR><TD>%p<TD>Directory (virtual path)
<TR><TD>%r<TD>Relative path
<TR><TD>%s<TD>Size in decimal 

</TABLE>
</CENTER>

<H5>
	ListBottom
</H5>
Determines the output generated after iterating through files. See
'FilePattern' for a description of parameters.

<H5>
	MIMEIcon
</H5>
Specify a mapping from a MIME type to an icons relative path. The MIME type
is considered the first part of the specified string up the the first
whitespace. The MIME pattern is a simple regular expression which matches
wildcards (*). 
This directive
can be repeated multiple times to denote different MIME mappings. Typically
the pattern '*' would be mapped to a default icon.

<H5>
	DescriptionFile
</H5>
Specifies a file with descriptions of other files. Each line in the
file is of the form
<PRE>
	filename|its description
</PRE>
Whitespace will not be stripped out of the filename or description.
Typically this file would also be excluded from the directory list by
naming it in an 'Exclude' directive.
If present a files description can be accessed via parameter %d.

<H5>
	Include
</H5>
Specifies the file patterns which will included in the directory listing. The
pattern is a simple regular expression (allows wildcards (*)). The directory
indexer first allows file then excludes a subset of them with the 
'Exclude' directive. The include directive may be repeated multiple times.

<H5>
	Exclude
</H5>
Specifies the file patterns which will exclude from the directory listing. The
pattern is a simple regular expression (allows wildcards (*)). The directory
indexer first allows files using 'Include' patterns  then excludes a subset of
them. The exclude directive may be repeated multiple times.

<H5>
	LastModifiedFormat
</H5>
Specifies the format (according to the C library strftime() function) for
the last modified date of the file. The directive enables parameter
%l to contain this format.

<H5>
	HTTPEquiv
</H5>
Specifies an HTTP header to be sent to the remote client as part of the
response. This directive may be used to override the default response
media of 'text/html'. This pattern may be repeated multiple times.

<H5>
	SwapFileName
</H5>
If the current file matches the specified filename the replacement filename
will be subtituted. SwapFileName takes a string as argument, the pattern up
to the first whitespace character is considered the 'old' filename, subsequent
characters, excluding leading whitespace are considered the 'new' filename.

<H5>
	NewPattern
</H5>
The pattern must be a wellformed HTML fragment. It replaces the %m parameter
if the option "NewTag" is set and the file is newer than the specified
NewTreshold.

<H5>
	NewThreshold
</H5>
This parameter specifies the number of days, within a file or directory is
considered as new and tagged accordingly.

<H5>
	Options
</H5>
Specifies options which effect the behaviour of directory indexing, many
available facilities are optional for performance reasons, the table below
lists the available option flags with thier meanings.
<CENTER>
<TABLE BORDER=1>
<TH>Option
<TH>Meaning
<TR>

<TD>Name
<TD>Lookup the name of the file or directory and put it in parameter %n

<TR>
<TD>AltName
<TD>Lookup the alternate name of the file or directory and put it in 
parameter %a.

<TR>
<TD>Icon
<TD>Evaluate the appropriate icon file for the file based on its
extension and MIME type and make it available through parameter %i

<TR>
<TD>MIMEType
<TD>Record the MIME type of the current file in parameter %c.

<TR>
<TD>RelPath
<TD>Record the relative path (path component of URL) to the file in parameter
%r.

<TR>
<TD>Size
<TD>Record the filesize of the current file in parameter %s in full decimal.
If the current file is a directory '-' is placed in %s.
If more than one of 'Size', 'AbbrevSize' or 'FormatSize' are used the
results are undefined.

<TR>
<TD>AbbrevSize
<TD>The filesize is abbreviated and placed in parameter %s. The abbreviated
file size has the following form '<1K', '3M', '2K', etc.
If more than one of 'Size', 'AbbrevSize' or 'FormatSize' are used the
results are undefined.

<TR>
<TD>FormatSize
<TD>The filesize is formatted with the comma (',') character seperating
thousands. The formatted size of place in parameter $s.
If more than one of 'Size', 'AbbrevSize' or 'FormatSize' are used the
results are undefined.

<TR>
<TD>SortDesc
<TD>The sort order is alphabetical descending. If this option is omitted
the sort order is ascending by default.

<TR>
<TD>SortCase
<TD>The sort order is case sensitive. If this option is omitted
the sort order is case insensitive by default.

<TR>
<TD>SortType
<TD>The sort order is determined by the file extension. Directories are
on the top of the sort order.

<TR>
<TD>SortName
<TD>The sort order is determined by the name of the file or directory.

<TR>
<TD>SortSize
<TD>The sort order is determined by the filesize.

<TR>
<TD>SortDate
<TD>The sort order is determined by the file date.

<TR>
<TD>SortDscr
<TD>The sort order is determined by the annotations in the DescriptionFile.

<TR>
<TD>TagNew
<TD>All all files newer than NewThreshold are marked with NewPattern.

</TABLE>
</CENTER>
Multiple options may be listed together by seperating them using the pipe
('|') character. Whitespace will be stripped from the option pattern. Options
are matched without regard to case sensitivity. Unknown options will raise
a configuration error. The option directive may be repeated multiple times.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED if a directory index was succesfully sent to the client.
	PIAPI_CONTINUE if this handler done nothing. PIAPI_ERROR and PIAPI_ABORT
	on generic and severe error conditions.

Note:
	A sorted HTML directory indes is made by using the above sort options and
	hyperlinks with special URL parameters within the table headers

	<XMP>
	...
	<TH><A HREF="%p?SortType=A">Type</A> \
	<TH><A HREF="%p?SortName=A">Name</A> \
	<TH><A HREF="%p?SortSize=A">Size</A> \
	<TH><A HREF="%p?SortDate=A">Last Modified</A> \
	<TH><A HREF="%p?SortDscr=A">Description</A> \
	...
	</XMP>

	The URL parameters are of the general form "SortType(Name,Size,Date,Dscr)=A(D)"
	for <B>A</B>scending or <B>D</B>escending sort order.

Example:
	<PRE>
	&lt;Object&gt;
		Name DirectoryIndex
		Class DirectoryIndexClass
		Options "Name | Size | MIMEType"
	&lt;/Object&gt;

	&lt;Object&gt;
		...
		Handle DirectoryIndex 
		...
	&lt;/Object&gt;
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif


/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Parse, store and add RFC822 headers to  response DB's.
\*____________________________________________________________________________*/
class RFC822Header
{
private:
	const char *pFKVariable;
	PIString sValue;
	int iOK;

public:
	RFC822Header( const char *pLine )
	:   pFKVariable( 0 ),
		iOK( 0 )
		{
		if ( !pLine )
			{ return; };
		int iLen = strlen( pLine );
		int i;
		for( i=0; i<iLen && pLine[i]!=':' && !(isspace(pLine[i])); i++);
		if ( !i || pLine[i]!=':' )
			{ return; };
		PIString sVariable( pLine, i );
		for( i++; i<iLen && (isspace(pLine[i])); i++);  /* scan past ws */
		pFKVariable = PIDB_getFastKey( sVariable, PIDBTYPE_RFC822 );
		sValue = &( pLine[i] );
		iOK = 1;
		};

	/* --- Replace this header in DB --- */
	void ReplaceHeader( PIDB *pDB )
		{
		assert( pDB );
		assert( pFKVariable );
		assert( iOK );
		PIDB_replace( pDB, PIDBTYPE_RFC822, pFKVariable,
			(void *)(const char *)sValue, PIDBFLAG_FASTKEY );
		};

	inline int IsOK()
		{ return iOK; };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Maps a MIME type to an image path.
\*____________________________________________________________________________*/
class MIMEIcon
{
private:
	PIString sMIMEPattern;
	PIString sIconPath;

public:
	MIMEIcon( const char *pMIMEPattern, const char *pIconPath )
	:	sMIMEPattern( pMIMEPattern ), sIconPath( pIconPath )
		{};

	/*___
	____*/
	bool Matches( const PIString &sMIMEType )
		{
		return HTTPUtil_regexMatch(
				sMIMEPattern, sMIMEPattern.Len(),
				sMIMEType, sMIMEType.Len() ) ?
			true : false;
		};

	/*___
	____*/
	inline const PIString &GetPath()	{ return sIconPath; };
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class SwapFileName
{
public:
	PIString sFrom;
	PIString sTo;

	SwapFileName( const PIString &sTheFrom, const PIString &sTheTo )
		{
		DeQuote tFrom( sTheFrom );
		DeQuote tTo( sTheTo );
		sFrom = (const char *)tFrom;
		sTo = (const char *)tTo;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class DescriptionMap
{
public:
	PIString sFileName;
	PIString sDescription;

	DescriptionMap( const char *pLine )
		{
		assert( pLine );
		int iLen = strlen( pLine );
		int i=0;
		for( i=0; i<iLen && pLine[i]!='|'; i++ );
		PIString sTmp( pLine, i );
		sFileName = sTmp;
		if ( i<iLen ) { i++; };
		pLine= &( pLine[i] );
		sDescription = pLine;
		};
};

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
				WriteContext for parameters and 
			callback functions for writing the parameters

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
/* --- helper macro to facilitate writing --- */
inline int Internal_StrMinCpy( char *pS1, const char *pS2, int iL1, int iL2 )
{
	if ( !iL2 )
		{ return 0; };
	if ( iL1 )
		{
		assert( iL1>0 );
		assert( pS1 );
		assert( pS2 );
		strncpy( pS1, pS2, (iL1<iL2) ? iL1 : iL2 );
		return iL2;
		};
	return iL2;
};
inline void _myassert( void *pVoid )
	{ assert( pVoid ); };
#define FUNCTION(fn,record) \
static int (fn)( PIHTTP *, void *pData, char *pszBuffer, int iLength ) { \
	_myassert( pData );										\
	return Internal_StrMinCpy(								\
		pszBuffer,											\
		((WriteContext *)pData)->record.pField,				\
		iLength,											\
		((WriteContext *)pData)->record.iLength ); }
class WriteContext
{
public:
	WriteContext()
	{
	memset( this, 0, sizeof( WriteContext ) );
	};
	class Record {
	public:
		int iLength;
		const char *pField;
	};
	Record tAltName;
	Record tIsDirectory;
	Record tMedia;
	Record tDescription;
	Record tLastModifiedDate;
	Record tName;
	Record tIcon;
	Record tDirectory;
	Record tRelativePath;
	Record tSize;
	Record tTagNew;
};
FUNCTION( f_a, tAltName );
FUNCTION( f_b, tIsDirectory );
FUNCTION( f_c, tMedia );
FUNCTION( f_d, tDescription );
FUNCTION( f_i, tIcon );
FUNCTION( f_l, tLastModifiedDate );
FUNCTION( f_n, tName );
FUNCTION( f_p, tDirectory );
FUNCTION( f_r, tRelativePath );
FUNCTION( f_s, tSize );
FUNCTION( f_m, tTagNew );
	
/*____________________________________________________________________________*\
 *
 Description:
	Create a map of FnPi3Write functions for parameters to expressions
\*____________________________________________________________________________*/
static FnPi3Write aFunctions[256] =
{
/*
0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
*/
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 0 .. 15 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 16 .. 31 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 32 .. 47 */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 48 .. 63 */
/*  A   B   C   D   E   F   G   H   I   J   K   L   M   N   O  */
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  /* 64 .. 79 */	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	/* 80 .. 95 */
/*  a   b   c   d   e   f   g   h   i   j   k   l   m   n   o  */
0,f_a,f_b,f_c,f_d,  0,  0,  0,  0,f_i,  0,  0,f_l,f_m,f_n,  0, 	/* 96 .. 111 */
/*  q   r   s   t   u   v   w   x   y   z   */
f_p,0,f_r,f_s,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  /* 112 .. 127 */	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 	
0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 	/* .. 255 */	
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_HEADERPATTERN		"HeaderPattern"
#define KEY_CONF_HEADERFILE			"HeaderFile"
#define KEY_CONF_LISTTOP			"ListTop"
#define KEY_CONF_FILEPATTERN		"FilePattern"
#define KEY_CONF_LISTBOTTOM			"ListBottom"
#define KEY_CONF_FOOTERFILE			"FooterFile"
#define KEY_CONF_FOOTERPATTERN		"FooterPattern"
#define KEY_CONF_LASTMODIFIEDFORMAT	"LastModifiedFormat"
#define KEY_CONF_INCLUDE			"Include"
#define KEY_CONF_EXCLUDE			"Exclude"
#define KEY_CONF_DESCRIPTIONFILE	"DescriptionFile"
#define KEY_CONF_MIMEICON			"MIMEIcon"
#define KEY_CONF_HTTPEQUIV			"HTTPEquiv"
#define KEY_CONF_SWAPFILENAME		"SwapFileName"
#define KEY_CONF_OPTIONS			"Options"
#define KEY_CONF_DESCRIPTIONFILE	"DescriptionFile"
#define KEY_CONF_NEWPATTERN			"NewPattern"
#define KEY_CONF_NEWTHRESHOLD		"NewTreshold"

/*____________________________________________________________________________*\
 *
 Flags and Options
\*____________________________________________________________________________*/
#define FLG_NONE		0x00000000
#define FLG_NAME		0x00000001
#define FLG_ALTNAME		0x00000002	
#define FLG_ICON		0x00000004
#define FLG_SIZE		0x00000008
#define FLG_MIMETYPE	0x00000010
#define FLG_RELPATH		0x00000020
#define FLG_MODDATE		0x00000040
#define FLG_ABBREVSIZE	0x00000080
#define FLG_FORMATSIZE	0x00000100
#define FLG_SORTCASE	0x00000200
#define FLG_SORTDESC	0x00000400
#define FLG_SORTTYPE	0x00000800
#define FLG_SORTNAME	0x00001000
#define FLG_SORTSIZE	0x00002000
#define FLG_SORTDATE	0x00004000
#define FLG_SORTDSCR	0x00008000
#define FLG_TAGNEW		0x00010000

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
				Map these flags to names

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
struct {
	const char *pName;
	int iFlag;
} aFlagMap[] =
	{
	{ "Name",					FLG_NAME },
	{ "AltName",				FLG_ALTNAME },
	{ "Icon",					FLG_ICON },
	{ "Size",					FLG_SIZE },
	{ "MIMEType",				FLG_MIMETYPE },
	{ "RelPath",				FLG_RELPATH },
	{ "AbbrevSize",				FLG_ABBREVSIZE },
	{ "FormatSize",				FLG_FORMATSIZE },
	{ "SortCase",				FLG_SORTCASE },
	{ "SortDesc",				FLG_SORTDESC },
	{ "SortType",				FLG_SORTTYPE },
	{ "SortName",				FLG_SORTNAME },
	{ "SortSize",				FLG_SORTSIZE },
	{ "SortDate",				FLG_SORTDATE },
	{ "SortDscr",				FLG_SORTDSCR },
	{ "TagNew",					FLG_TAGNEW },
	/* ---
	end of change
	--- */

	/* --- leave this last always --- */
	{ 0, 0 }
	};

#define KEY_SORT_TYPE	"SortType"
#define KEY_SORT_NAME	"SortName"
#define KEY_SORT_SIZE	"SortSize"
#define KEY_SORT_DATE	"SortDate"
#define KEY_SORT_DESC	"SortDscr"
#define SORT_KEY_LEN    8

#define SORT_NONE	0
#define SORT_TYPE	1
#define SORT_NAME	2
#define SORT_SIZE	3
#define SORT_DATE	4
#define SORT_DESC	5
#define MAX_FILESORT_ARRAY	2048
/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Helper function, compare two keys of type PIFInfo * by directory property.
	Trick: Result is emphasized by 2 --> 0 no directory | >0 compare test
\*____________________________________________________________________________*/

int cmpDirs( PIFInfo *arg1, PIFInfo *arg2 ) {
	if ( PIFInfo_isDirectory(arg1) ) {
		// the parent dir has always the lowest weight
		if ( !strcmp( PIFInfo_getName(arg1), "..\0" )) return 1;
		if ( PIFInfo_isDirectory(arg2) ) {
			// the parent dir has always the lowest weight
			// but here it's the 2nd compare parameter
			if ( !strcmp( PIFInfo_getName(arg2), "..\0" )) return 3;
			// both directories --> equal weight
			return 2;
		};
		// default case if a1 is dir
		return 1;
	} else {
		// default case if a2 is dir
		if ( PIFInfo_isDirectory(arg2)) return 3;
		// default case
		return 0;
	};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Compare two keys of type PIFInfo * by type.
\*____________________________________________________________________________*/

int cmpType( PIFInfo *arg1, PIFInfo *arg2, PIDB *pDB, int bDescending, int bCaseSensitive ) {
	// assume directories has 0 bytes (smaller than files)
	// directories are always top level
	int result = 0;
	int Sort = cmpDirs( arg1, arg2 );
	if ( Sort ) {
		result = Sort - 2;
	} else {
		// case sensitivity
		if ( bCaseSensitive ) {
			result = strcmp(HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg1 )),
				HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg2 )));
		} else {
			result = PIUtil_stricmp(HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg1 )),
				HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg2 )));
		};
	};

	// sort order
	return bDescending ? -result : result;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Compare two keys of type PIFInfo * by name.
\*____________________________________________________________________________*/

int cmpName( PIFInfo *arg1, PIFInfo *arg2, PIDB *pDB, int bDescending, int bCaseSensitive ) {
	// case sensitivity
	int result = bCaseSensitive ? strcmp( PIFInfo_getName( arg1 ), PIFInfo_getName( arg2 ))
		: PIUtil_stricmp( PIFInfo_getName( arg1 ), PIFInfo_getName( arg2 ));
	// sort order
	return bDescending ? -result : result;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Compare two keys of type PIFInfo * by size.
\*____________________________________________________________________________*/

int cmpSize( PIFInfo *arg1, PIFInfo *arg2, PIDB *pDB, int bDescending, int bCaseSensitive ) {
	// assume directories has 0 bytes (smaller than files)
	// directories are always top level
	int Sort = cmpDirs( arg1, arg2 );
	int result = ( Sort ) ? Sort - 2 : PIFInfo_getSize( arg1 ) - PIFInfo_getSize( arg2 );
	// no case sensitivity
	// sort order
	return bDescending ? -result : result;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Compare two keys of type PIFInfo * by date.
\*____________________________________________________________________________*/

int cmpDate( PIFInfo *arg1, PIFInfo *arg2, PIDB *pDB, int bDescending, int bCaseSensitive ) {
	int result = 0;
	if ( PIFInfo_getLastModified( arg1 ) > PIFInfo_getLastModified( arg2 ))
		{ result = 1;	}
	else {
		if ( PIFInfo_getLastModified( arg1 ) < PIFInfo_getLastModified( arg2 ))
			{ result = -1; }
		else
			{ result = 0; };
	};
	// no case sensitivity
	// sort order
	return bDescending ? -result : result;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Compare two keys of type PIFInfo * by description.
\*____________________________________________________________________________*/

int cmpDesc( PIFInfo *arg1, PIFInfo *arg2, PIDB *pDB, int bDescending, int bCaseSensitive) {
	// load default description (Mime-Type) for 1st argument
	const char *pVal1 = PIFInfo_isDirectory( arg1 ) ? TYPE_MAGIC_DIRECTORY
		:HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg1 ));

	// load default description (Mime-Type) for 2nd argument
	const char *pVal2 = PIFInfo_isDirectory( arg2 ) ? TYPE_MAGIC_DIRECTORY
		:HTTPCore_getMIMETypeFromExtension( PIFInfo_getExtension( arg2 ));

	if ( pDB ) {
		// load 1st description from Map-DB
		DescriptionMap *pDesc1 = (DescriptionMap *)PIDB_lookup(
			pDB, PIDBTYPE_OPAQUE, PIFInfo_getName( arg1 ), 0 );
		if ( pDesc1 ) pVal1 = pDesc1->sDescription;

		// load 2nd description from Map-DB
		DescriptionMap *pDesc2 = (DescriptionMap *)PIDB_lookup(
			pDB, PIDBTYPE_OPAQUE, PIFInfo_getName( arg2 ), 0 );
		if ( pDesc2 ) pVal2 = pDesc2->sDescription;
	};
	// description could contain HTML-tags (not visible but compared)
	if (strchr(pVal1, '<')==pVal1) { pVal1 = strchr(pVal1, '>'); pVal1++; }
	if (strchr(pVal2, '<')==pVal2) { pVal2 = strchr(pVal2, '>'); pVal2++; }
	// case sensitivity
	int result = bCaseSensitive ? strcmp( pVal1, pVal2 ) : PIUtil_stricmp( pVal1, pVal2 );
	// sort order
	return bDescending ? - result : result;
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Quick sort of an array of PIFInfo pointers, by given column no.
	Optional descending, case sensitive sort flags.
\*____________________________________________________________________________*/

void qSort(PIFInfo *A[], int Min, int Max, PIDB *pDB, int iSortByColumn, int bDescending, int bCaseSensitive) {
	if ( iSortByColumn == SORT_NONE) { return; };
	int Lo = Min;
	int Hi = Max;
	PIFInfo *T;
	PIFInfo *Mid = A[(Lo + Hi) / 2];
    do {
		switch ( iSortByColumn )
		{
			case SORT_TYPE:
				while ( cmpType( A[Lo], Mid, pDB, bDescending, bCaseSensitive) < 0 ) { Lo++; };
				while ( cmpType( A[Hi], Mid, pDB, bDescending, bCaseSensitive) > 0 ) { Hi--; };
				break;
			case SORT_NAME:
				while ( cmpName( A[Lo], Mid, pDB, bDescending, bCaseSensitive) < 0 ) { Lo++; };
				while ( cmpName( A[Hi], Mid, pDB, bDescending, bCaseSensitive) > 0 ) { Hi--; };
				break;
			case SORT_SIZE:
				while ( cmpSize( A[Lo], Mid, pDB, bDescending, bCaseSensitive) < 0 ) { Lo++; };
				while ( cmpSize( A[Hi], Mid, pDB, bDescending, bCaseSensitive) > 0 ) { Hi--; };
				break;
			case SORT_DATE:
				while ( cmpDate( A[Lo], Mid, pDB, bDescending, bCaseSensitive) < 0 ) { Lo++; };
				while ( cmpDate( A[Hi], Mid, pDB, bDescending, bCaseSensitive) > 0 ) { Hi--; };
				break;
			case SORT_DESC:
				while ( cmpDesc( A[Lo], Mid, pDB, bDescending, bCaseSensitive) < 0 ) { Lo++; };
				while ( cmpDesc( A[Hi], Mid, pDB, bDescending, bCaseSensitive) > 0 ) { Hi--; };
				break;
			default: break;
		};

		if ( Lo <= Hi ) {
	        T = A[Lo];
		    A[Lo] = A[Hi];
			A[Hi] = T;
			Lo++;
			Hi--;
		};
	} while ( Lo <= Hi );
	if ( Hi > Min ) { qSort( A, Min, Hi, pDB, iSortByColumn, bDescending, bCaseSensitive ); };
	if ( Lo < Max ) { qSort( A, Lo, Max, pDB, iSortByColumn, bDescending, bCaseSensitive ); };
};

/* ---
end of change
--- */

/*____________________________________________________________________________*\
 *
 Class:
 Description:
	Directory index handler
\*____________________________________________________________________________*/
class DirectoryIndex : public HandlerBaseHTTP
{
private:
	/* --- forbid copy constructor --- */
	DirectoryIndex( const DirectoryIndex &t )
	: HandlerBaseHTTP( t )
		{ assert( 0 ); };

	/* ---
	Configuration data
	--- */
	Pi3Expression *pHeaderPattern;	/* pattern for top */
	Pi3Expression *pListTopPattern;	/* pattern sent before index */
	Pi3Expression *pFilePattern;	/* pattern for file */
	Pi3Expression *pListBottomPattern;/* pattern sent after index */
	Pi3Expression *pFooterPattern;	/* pattern for footer */
	PIString sHeaderFile;			/* file to be sent as header */
	PIString sFooterFile;			/* file to be sent as footer */
	PIString sLastModifiedFormat;	/* last modified format */
	PIString sNewPattern;			/* pattern to tag new files */
	DblList lInclude;				/* patterns to include */
	DblList lExclude;				/* patterns to exclude */
	DblList lMIMEIcons;				/* mappings of MIME types to images */
	DblList lSwapFileNames;			/* files to swap */
	DblList lHTTPEquivalents;		/* response headers to send */
	PIString sDescriptionFile;		/* description file */
	int iOptions;					/* options/flags */
	time_t tNewTreshold;			/* threshold time for new files */

	/* ---
	FastKeys
	--- */
	const char *pFKContentType;			/* RFC822: Content-Type */
	const char *pFKURI;					/* RFC822: URI from request header */
	const char *pFKPath;				/* STRING: Translated path */

	PIString sListTop;
	int bSortDescending;
	int iSortByColumn;
	/* ---
	end of change
	--- */
	
protected:
	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

				Load an expression or condition

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int LoadExpression( const char *pName, Pi3Expression **ppExpression,
		FnPi3Write *pFunctions, const char *pValue, PIOStrStream &os )	
		{
		assert( ppExpression );
		Pi3Expression *pExpression = *ppExpression;
		if ( pExpression )
			{
			os << "'" << pName << "' may only be specified once." << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};
		Pi3String *pError = Pi3String_new( 0 );
		pExpression = Pi3Expression_new( pValue, pFunctions, pError );
		if ( !pExpression )
			{
			os << "Error parsing expression: " <<
				Pi3String_getPtr( pError ) << ends;
			CONFIG_ERR( Object(), os.str() );
			Pi3String_delete( pError );
			return 0;
			};
		Pi3String_delete( pError );
		*ppExpression = pExpression;
		return 1;
		};

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

		Configure handler object based on parameters

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "DirectoryIndex: ";
		if ( !PIUtil_stricmp( KEY_CONF_HEADERFILE, pVariable ) )
			{
			sHeaderFile = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FOOTERFILE, pVariable ) )
			{
			sFooterFile = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_LASTMODIFIEDFORMAT, pVariable ) )
			{
			iOptions |= FLG_MODDATE;
			sLastModifiedFormat = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_HEADERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_HEADERPATTERN, &pHeaderPattern,
				aFunctions, pValue, os ) )
				{ return 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_LISTTOP, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_LISTTOP, &pListTopPattern,
				aFunctions, pValue, os ) )
				{ return 0;	};
			sListTop = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FILEPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_FILEPATTERN, &pFilePattern,
				aFunctions, pValue, os ) )
				{ return 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FOOTERPATTERN, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_FOOTERPATTERN, &pFooterPattern,
				aFunctions, pValue, os ) )
				{ return 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_LISTBOTTOM, pVariable ) )
			{
			if ( !LoadExpression( KEY_CONF_LISTBOTTOM, &pListBottomPattern,
				aFunctions, pValue, os ) )
				{ return 0; };
			}
		else if ( !PIUtil_stricmp( KEY_CONF_OPTIONS, pVariable ) )
			{
			/* ---
			Split the options up by the bar ('|') character and then
			strip trailing and leading whitespace from option before
			attempting to match it.
			--- */
			StringTokenizer tTokens( pValue, "|" );
			for(int i=0; i<tTokens.NumTokens(); i++)
				{
				const char *pToken = tTokens.GetToken( i );

				/* --- skip leading whitespace --- */
				for( ; *pToken && (isspace(*pToken)); pToken++ );

				/* ---
				include only to first whitespace trailing whitespace
				--- */
				int j=0;
				for( ; pToken[j] && !(isspace(pToken[j])); j++ );
				/* ---
				j now contains the length of the first word in pToken
				--- */

				/* ---
				cycle through the list of available flags
				comparing them with this one
				--- */
				int i=0;
				for( ; aFlagMap[i].pName; i++ )
					{
					if ( !PIUtil_strncmpi( aFlagMap[i].pName, pToken, j ) )
						{ break; };
					};
				if ( !aFlagMap[i].pName )
					{
					/* --- flag not found --- */
					PIString sTmp( pToken, j );
					os << "Unknown option flag '" << sTmp << "'." << ends;
					CONFIG_ERR( Object(), os.str() );
					return 0;
					};
				iOptions |= aFlagMap[i].iFlag;
				};	/* --- loop over tokens seperated by '|' --- */
			}
		else if ( !PIUtil_stricmp( KEY_CONF_INCLUDE, pVariable ) )
			{
			lInclude.Append( (DblList::type) PI_NEW( PIString( pValue ) ) );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_EXCLUDE, pVariable ) )
			{
			lExclude.Append( (DblList::type) PI_NEW( PIString( pValue ) ) );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_SWAPFILENAME, pVariable ) )
			{
			int i;
			for( i=0; pValue[i] && !(isspace(pValue[i])); i++);
			PIString sOld( pValue, i );
			pValue = &( pValue[i] );
			for( ; *pValue && (isspace(*pValue)); pValue++ );
			lSwapFileNames.Append( (DblList::type)
				PI_NEW( SwapFileName( sOld, pValue ) ) );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_DESCRIPTIONFILE, pVariable ) )
			{
			sDescriptionFile = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_MIMEICON, pVariable ) )
			{
			StringTokenizer tTokens( pValue, " ", 0 );
			if ( tTokens.NumTokens()!=2 )
				{	
				os << "Syntax error in definiton of 'MIMEIcon', line is '" <<
					pValue << "'" << ends;
				CONFIG_ERR( Object(), os.str() );
				return 0;
				};
			lMIMEIcons.Append( (DblList::type) PI_NEW(
				MIMEIcon( tTokens.GetToken(0), tTokens.GetToken(1) ) ) );
			}
        else if ( !PIUtil_stricmp( KEY_CONF_HTTPEQUIV, pVariable ) )
            {
            RFC822Header *pHeader = PI_NEW( RFC822Header( pValue ) );
            if ( !pHeader || !pHeader->IsOK() )
                {
                os << "Bad HTTPEquiv expression, '" << pValue << "'" << ends;
                CONFIG_ERR( Object(), os.str() );
                PI_DELETE( pHeader );
                return 0;
                }
            else
                {
                lHTTPEquivalents.Append( (DblList::type)pHeader );
                };
            }
		else if ( !PIUtil_stricmp( KEY_CONF_NEWPATTERN, pVariable ) )
			{
			sNewPattern = pValue;
			}
		else if ( !PIUtil_stricmp( KEY_CONF_NEWTHRESHOLD, pVariable ) )
			{
			tNewTreshold = 86400 * atoi(pValue);
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
				Object lifetime stuff

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	DirectoryIndex( PIObject *pTheObject, int iArgc, const char *ppArgv[] )
	:
		HandlerBaseHTTP( pTheObject ),
		pHeaderPattern( 0 ),
		pListTopPattern( 0 ),
		pFilePattern( 0 ),
		pListBottomPattern( 0 ),
		pFooterPattern( 0 ),
		iOptions( FLG_NONE ),
		pFKContentType( PIDB_getFastKey( KEY_HTTP_CONTENTTYPE,PIDBTYPE_RFC822)),
		pFKURI( PIDB_getFastKey( KEY_HTTP_URI, PIDBTYPE_STRING ) ),
		pFKPath( PIDB_getFastKey( KEY_INT_PATH, PIDBTYPE_STRING ) ),
		sListTop( 0 ),
		bSortDescending( 0 ),
		iSortByColumn( SORT_NONE ),
		tNewTreshold( 0 )
		{
		ReadParameters( iArgc, ppArgv );
		if ( !IsOK() )
			{ return; };

		/* --- validate all necessary components were loaded --- */
		if ( !pFilePattern )
			{
			CONFIG_ERR( Object(),"DirectoryIndex: 'FilePattern' not defined");
			SetOK( 0 );
			return;
			};
		};

	virtual ~DirectoryIndex()
		{
		for( DblListIterator i( lInclude ); !i.BadIndex(); i++)
			{ PI_DELETE( (PIString *)i.Current() ); };
		for( DblListIterator j( lExclude ); !j.BadIndex(); j++)
			{ PI_DELETE( (PIString *)j.Current() ); };
		for( DblListIterator k( lMIMEIcons ); !k.BadIndex(); k++)
			{ PI_DELETE( (MIMEIcon *)k.Current() ); };
		for( DblListIterator m( lSwapFileNames ); !m.BadIndex(); m++)
			{ PI_DELETE( (SwapFileName *)m.Current() ); };
        for( DblListIterator n( lHTTPEquivalents ); !n.BadIndex(); n++)
            { PI_DELETE( (RFC822Header *)n.Current() ); };
		Pi3Expression_delete( pHeaderPattern );
		Pi3Expression_delete( pListTopPattern );
		Pi3Expression_delete( pFilePattern );
		Pi3Expression_delete( pListBottomPattern );
		Pi3Expression_delete( pFooterPattern );
		};

private:
	/* --- internal methods used to write directory index --- */

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

			Miscellanious utilities

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	/* ---
	Check if a filename should be changed to a different filename, returns
	new filename or PIString::Empty() to indicate no swap.
	--- */
	const PIString &CheckForSwap( const char *pFileName ) const
		{
		for( ConstDblListIterator i( lSwapFileNames ); !i.BadIndex(); i++ )
			{
			SwapFileName &tSwap = *((SwapFileName *)i.Current());
			if ( !PIUtil_stricmp( tSwap.sFrom, pFileName ) )
				{
				return tSwap.sTo;
				};
			};
		return PIString::Empty();
		};

	/* ---
	Returns the Icon image for a MIME type
	--- */
	const PIString &GetMIMEIcon( const PIString &sMIME ) const
		{
		for( ConstDblListIterator i( lMIMEIcons ); !i.BadIndex(); i++ )
			{
			MIMEIcon &tIcon = *( (MIMEIcon *)i.Current() );
			if (  tIcon.Matches( sMIME ) )
				{ return tIcon.GetPath(); };
			};
		return PIString::Empty();
		};

	/* --- --- */
	bool DoWriteThousands( PIString &sNum, unsigned long ulVal,
		const char *pSep ) const
		{
		if ( ulVal )
			{
			enum { BUF_SIZE=16 };
			char szBuf[BUF_SIZE];
			*szBuf='\0';
			if ( DoWriteThousands( sNum, ulVal/1000, pSep ) )
				{
				sNum.Concatenate( pSep );
				sprintf( szBuf, "%03d", (int)(ulVal%1000) );
				}
			else
				{
				sprintf( szBuf, "%d", (int)(ulVal%1000) );
				};
			sNum.Concatenate( szBuf );
			return true;
			}
		else
			{ return false; };
		};

	/*___
	____*/
	bool WriteThousands( PIString &sNum, unsigned long ulVal,
		const char *pSep ) const
		{
		if ( ulVal )
			{
			sNum = PIString::Empty();
			return DoWriteThousands( sNum, ulVal, pSep );
			}
		else
			{ sNum = "0"; };
		return true;
		};

	/* ---
	Load a description file, if one exists and contains descriptions return 1
	else 0
	--- */
	int LoadDescriptionFile( PIDB *pDescriptionDB, PIFInfo *pDir ) const
		{
		assert( pDir );
		int iRet = 0;
		if ( sDescriptionFile==PIString::Empty() )
			{ /* no file specified */ return iRet; };
		PIString sPath( PIFInfo_getPath( pDir ) );
		sPath.Concatenate( "/" );
		sPath.Concatenate( sDescriptionFile );

		/* --- get the file --- */
		PIFInfo *pFInfo = HTTPCore_getCachedFile( sPath );
		if ( !pFInfo )
			{ /* error should always be able to get file object */ return 0; };

		if ( PIFInfo_exists( pFInfo )  && PIFInfo_isRegular( pFInfo ) )
			{	
			enum { BUF_SIZE=1023 };
			char szBuf[BUF_SIZE+1];
			*szBuf = '\0';
			ifstream ifs( PIFInfo_getPath( pFInfo ) );
			while( ifs && !ifs.eof() ) 	
				{
				ifs.getline( szBuf, BUF_SIZE );
				DescriptionMap *pDesc = PI_NEW( DescriptionMap( szBuf ) );
				PIDB_add( pDescriptionDB, PIDBTYPE_OPAQUE, pDesc->sFileName,
					(void *)pDesc, 0 );
				iRet = 1;
				};
			};
		HTTPCore_releaseCachedFile( pFInfo );
		return iRet;
		};

	/* ---
	send a header or footer file, if appropriate.
	returns 1 on OK, 0 on error.
	--- */
	int SendFile( PIHTTP &tPIHTTP, const PIString &sFile, PIFInfo *pDir ) const
		{
		assert( pDir );
		if ( sFile==PIString::Empty() )
			{ /* no file specified */ return 1; };
		PIString sPath( PIFInfo_getPath( pDir ) );
		sPath.Concatenate( "/" );
		sPath.Concatenate( sFile );

		/* --- get the file --- */
		PIFInfo *pFInfo = HTTPCore_getCachedFile( sPath );
		if ( !pFInfo )
			{ /* error should always be able to get file object */ return 0; };

		if ( PIFInfo_exists( pFInfo )  && PIFInfo_isRegular( pFInfo ) )
			{	
			/* --- memory map the file and send it --- */
			PIFMap *pFileMap = PIFMap_new( pFInfo );

			/* --- send mapped file --- */
			const char *pMap = 0;
			int iLen = 0;
			if ( pFileMap )
				{ pMap = (const char *)PIFMap_lock( pFileMap, &iLen ); };
	
			if ( pMap && iLen )
				{
				PIIOBuffer_write( tPIHTTP.pBuffer, pMap, iLen, PIIOBUF_NONE );
				};

			PIFMap_delete( pFileMap );
			};
		HTTPCore_releaseCachedFile( pFInfo );
		return 1;
		};

	/* ---
	Grap the IOBuffer's internal buffer pointer and write the
	expression directly into it
	--- */
	int WriteExpression( Pi3Expression *pExpr, PIHTTP &tPIHTTP,
		WriteContext &tContext ) const
		{
		if ( !pExpr )
			{ return 1; };
		assert( pExpr );
		int iRecursed = 0;
		for(;;)
			{
			char *pBuffer = 0;
			int iSize = PIIOBuffer_getOutputBuffer( tPIHTTP.pBuffer, &pBuffer );
			int iLen = Pi3Expression_write( pExpr, &tPIHTTP, &tContext,
				pBuffer, iSize );
			if ( !iLen )
				{ return 1; };
			if ( iSize<iLen )
				{
				if ( iRecursed )
					{
					HTTPCore_logError( &tPIHTTP, "DirectoryIndex: IO \
buffer to small to generate expression for output." );
					return 0;
					};
				PIIOBuffer_flush( tPIHTTP.pBuffer );
				iRecursed = 1;
				}
			else
				{
				PIIOBuffer_advanceBufferPointer( tPIHTTP.pBuffer, iLen );
				return 1;
				};
			};

		/* --- should never get here --- */
		assert( 0 );
		return 0;
		};

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

		Generate the header for the directory listing

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int Top( PIHTTP &tPIHTTP, PIFInfo *pDir, WriteContext &tContext ) const
		{
		if ( !WriteExpression( pHeaderPattern, tPIHTTP, tContext ) )
			{ return 0; };

		/* --- send any header file --- */
		if ( !SendFile( tPIHTTP, sHeaderFile, pDir ) )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Error \
sending header file '%s'.", (const char *)sHeaderFile );
			return 0;
			};

		if ( !WriteExpression( pListTopPattern, tPIHTTP, tContext ) )
			{ return 0; };

		return 1;
		}

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

		Generate a output for a single file in the listing

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int File( PIHTTP &tPIHTTP, WriteContext &tContext, PIFInfo *pInfo,
		const char *pRelativePath, PIDB *pDescriptionDB ) const
		{
		register int iFlags = iOptions;
		assert( pInfo );

		const char *pName = PIFInfo_getName( pInfo );
		assert( pName );
		if ( !pName ) { return 0; };

		/* --- name --- */
		if ( iFlags & FLG_NAME )
			{
			const PIString &sTmp = CheckForSwap( pName );
			if ( sTmp==PIString::Empty() )
				{
				tContext.tName.pField = PIFInfo_getName( pInfo );
				tContext.tName.iLength = strlen( PIFInfo_getName( pInfo ) );
				}
			else
				{
				tContext.tName.pField = sTmp;
				tContext.tName.iLength = sTmp.Len();
				};
			};

		/* --- alternate name --- */
		if ( iFlags & FLG_ALTNAME )
			{
			const char *pAlt = PIFInfo_getAlternateName( pInfo );
			tContext.tAltName.pField = pAlt;
			tContext.tAltName.iLength = strlen( pAlt );
			};

		/* --- modification date --- */
		enum { DATE_BUFFER=128 };
		char szDate[DATE_BUFFER];
		*szDate = '\0';
		if ( iFlags & FLG_MODDATE )
			{
			time_t tT = PIFInfo_getLastModified( pInfo );
			PIPlatform_beforeUnsafeBlock();
			struct tm *pTms = gmtime( &tT );
			assert( pTms );
			strftime( szDate, DATE_BUFFER, sLastModifiedFormat, pTms );
			PIPlatform_afterUnsafeBlock();
			tContext.tLastModifiedDate.pField = szDate;
			tContext.tLastModifiedDate.iLength = strlen( szDate );
			};

		/* --- tag file if new --- */
		if ( iFlags & FLG_TAGNEW )
			{
			time_t tT1 = PIFInfo_getLastModified( pInfo );
			time_t tT2;
			PIPlatform_beforeUnsafeBlock();
			time(&tT2);
			PIPlatform_afterUnsafeBlock();
			if ( tT1 + tNewTreshold >= tT2 )
				{
				tContext.tTagNew.pField = (const char *)sNewPattern;
				tContext.tTagNew.iLength = strlen( sNewPattern );
				}
			else
				{
				tContext.tTagNew.pField = 0;
				tContext.tTagNew.iLength = 0;
				}
			};

		/* ---
		Determine whether or not this is a directory
		--- */
		bool bIsDirectory = PIFInfo_isDirectory( pInfo );
		if ( bIsDirectory )
			{
			tContext.tIsDirectory.pField = "true";
			tContext.tIsDirectory.iLength = 4;
			}
		else
			{
			tContext.tIsDirectory.pField = 0;
			tContext.tIsDirectory.iLength = 0;
			};

		/* --- relative path --- */
		PIString sRelativePath( pRelativePath );
		if ( iFlags & FLG_RELPATH )
			{
			/* ---
			relative URI to file
			--- */
			/* --- don't do full URL encoding, just substitute '+' for ' '--- */
			PIString sName( pName );
			SwapSubStrings tEncoded( sName, " ", "+" );
			sRelativePath.Concatenate( tEncoded.GetString() );
			if ( bIsDirectory )
				{
				sRelativePath.Concatenate( HTTP_DIRSEPERATOR );
				};
			tContext.tRelativePath.pField = sRelativePath;
			tContext.tRelativePath.iLength = sRelativePath.Len();
			};

		PIString sSize;
		if ( ( iFlags & FLG_SIZE ) ||
			( iFlags & FLG_FORMATSIZE ) ||
			( iFlags & FLG_ABBREVSIZE )
			)
			{
			if ( bIsDirectory )
				{
				tContext.tSize.pField = "-";
				tContext.tSize.iLength = 1;
				}
			else
				{
				unsigned long ulSize = PIFInfo_getSize( pInfo );
				if ( iFlags & FLG_SIZE )
 					{
					enum { BUF_SIZE=63 };
					char szBuf[BUF_SIZE+1];
					*szBuf = '\0';
					sprintf( szBuf, "%lu", ulSize );
					sSize = szBuf;
					}
				if ( iFlags & FLG_FORMATSIZE )
					{
					sSize = "0";
					WriteThousands( sSize, ulSize, "," /* thousands delimiter */ );
					}
				else if ( iFlags & FLG_ABBREVSIZE )
					{
					enum { BUF_SIZE=63 };
					char szBuf[BUF_SIZE+1];
					*szBuf = '\0';
					char s='K';
					if ( ulSize>=0x100000 )		/* 1 Meg. */
						{
						s = 'M';
						ulSize = 0xFFFFF & (ulSize >> 20); 
						sprintf( szBuf, "%d", (int)ulSize );
						sSize = szBuf;
						}
/*
**					else if ( ulSize>=0x400 )	/.* 1 K *./
** disable the <1K display because of alignment problems
*/
					else 
						{
						ulSize = 0x3FF & (ulSize >> 10); 
						if ( ulSize==0 ) { ulSize=1; };
						sprintf( szBuf, "%d", (int)ulSize );
						sSize = szBuf;
						}
/*					else
**						{ sSize = "&lt;1"; };
*/
					sSize.Concatenate( s );
					};
	
				/* --- set size --- */
				tContext.tSize.pField = sSize;
				tContext.tSize.iLength = sSize.Len();
				};
			};

		/* ---
		get the Media type only if it will be used.
		--- */
		PIString sMIME;
		if ( ( iFlags & FLG_MIMETYPE ) || ( iFlags & FLG_ICON ) )
			{
			if ( bIsDirectory )
				{ sMIME = TYPE_MAGIC_DIRECTORY; }
			else
				{
				 /* --- attempt to get mime type --- */
				sMIME = HTTPCore_getMIMETypeFromExtension(
					PIFInfo_getExtension( pInfo ) );
				}
			};

		/* --- Set the Media type --- */
		if ( iFlags & FLG_MIMETYPE )
			{
			tContext.tMedia.pField = sMIME;
			tContext.tMedia.iLength = sMIME.Len();
			};
	
		/* --- Get the icon path for this MIME type --- */
		if ( iFlags & FLG_ICON )
			{
			const PIString &sIcon = GetMIMEIcon( sMIME );
			tContext.tIcon.pField = sIcon;
			tContext.tIcon.iLength = sIcon.Len();
			};

		/* --- get the files description --- */
		if ( pDescriptionDB )
			{
			DescriptionMap *pDesc = (DescriptionMap *)PIDB_lookup(
				pDescriptionDB, PIDBTYPE_OPAQUE, pName, 0 );
			if ( pDesc )
				{
				tContext.tDescription.pField = pDesc->sDescription;
				tContext.tDescription.iLength = pDesc->sDescription.Len();
				}
			else
				{	
				tContext.tDescription.pField = 0;
				tContext.tDescription.iLength = 0;
				};
			};

		/* --- generate the output and send it to the remote client --- */
		assert( pFilePattern );
		if ( !WriteExpression( pFilePattern, tPIHTTP, tContext ) )
			{ return 0; };

		/* --- null out some fields --- */
		tContext.tName.iLength = 0;
		tContext.tAltName.iLength = 0;
		tContext.tLastModifiedDate.iLength = 0;
		tContext.tIsDirectory.iLength = 0;
		tContext.tRelativePath.iLength = 0;
		tContext.tIcon.iLength = 0;
		tContext.tMedia.iLength = 0;
		tContext.tSize.iLength = 0;

		return 1;
		};

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *
		
				Generate the directory trailer

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int Bottom( PIHTTP &tPIHTTP, PIFInfo *pDir, WriteContext &tContext ) const
		{
		if ( !WriteExpression( pListBottomPattern, tPIHTTP, tContext ) )
			{ return 0; }

		/* --- send any footer file --- */
		if ( !SendFile( tPIHTTP, sFooterFile, pDir ) )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Error \
sending footer file '%s'.", (const char *)sFooterFile );
			return 0;
			};

		if ( !WriteExpression( pFooterPattern, tPIHTTP, tContext ) )
			{ return 0; }

		return 1;
		};

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

		Generate the directory and write in to output stream

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	bool GenerateDirectoryListing(
			PIHTTP &tPIHTTP,
			WriteContext &tContext,
			PIFInfo *pDir,
			const char *pRelativePath ) const
		{
		assert( pDir );
		assert( PIFInfo_isDirectory( pDir ) );

		tContext.tDirectory.pField = pRelativePath;
		tContext.tDirectory.iLength = strlen( pRelativePath );

		/* ---	
		Attempt to load a description file
		--- */
		PIDB *pDescriptionDB = 0;
		if ( sDescriptionFile!=PIString::Empty() )
			{
			pDescriptionDB = PIDB_new( 0, "File Descriptions" );
			if ( !LoadDescriptionFile( pDescriptionDB, pDir ) )
				{
				PIDB_delete( pDescriptionDB );
				pDescriptionDB = 0;
				};
			};

		Top( tPIHTTP, pDir, tContext );

		PIFInfo *pFile = PIFInfo_getFirstFileInDirectory( pDir );
		int iError = 0;		/* error flag */
		PIFInfo *Files[MAX_FILESORT_ARRAY];
		int fCount = 0;
		
		while( pFile && !iError && ( fCount < MAX_FILESORT_ARRAY ))
			{
			/* ---
			Check file is allowed, this is an OR function, if any allow		
			pattern matches then include this file
			--- */
			int iAllow = 0;
			if ( PIFInfo_isReadable( pFile ) )
				{
				/* ---
				Get filename
				--- */
				PIString sFileName( PIFInfo_getName( pFile ) );

				for( ConstDblListIterator i( lInclude ); !i.BadIndex(); i++ )
					{
					PIString &sPattern = *( (PIString *)i.Current() );	
					if ( HTTPUtil_regexMatch(
						sPattern, sPattern.Len(),
						sFileName, sFileName.Len() )
						)
						{
						iAllow = 1;
						break;
						};
					};
	
				if ( iAllow )
					{
					/* ---
					Check if the file has been excluded, this is an OR function,
					exclude if any exlusion matches
					--- */
					for( ConstDblListIterator e( lExclude ); !e.BadIndex(); e++ )
						{
						PIString &sPattern = *( (PIString *)e.Current() );	
						if ( HTTPUtil_regexMatch(
							sPattern, sPattern.Len(),
							sFileName, sFileName.Len() )
							)
							{
							iAllow = 0;	
							break;
							};
						};
					};
				};
	
			if ( iAllow )
				{
				Files[fCount++] = PIFInfo_new( PIFInfo_getPath( pFile ));
				};

			if ( iError || !PIFInfo_getNextFileInDirectory( pFile ) )
				{ PIFInfo_delete( pFile ); pFile = 0; };
			}; //while

		if ( !iError )
			{
			Files[fCount] = 0;
			qSort(Files, 0, fCount - 1, pDescriptionDB, iSortByColumn, bSortDescending, iOptions & FLG_SORTCASE);
			}

		for (int j = 0; j < fCount; j++ ) {
			// call 'File' only if no error but delete array members in any case
			iError = iError || !File( tPIHTTP, tContext, Files[j],
				pRelativePath, pDescriptionDB );
			PIFInfo_delete( Files[j] );
		};

		Bottom( tPIHTTP, pDir, tContext );

		/* ---
		Delete descriptions 
		--- */
		PIDBIterator *pIter = PIDB_getIterator( pDescriptionDB,
			PIDBTYPE_OPAQUE, 0, 0 );
		if ( pIter )
			{
			for( ; pIter && PIDBIterator_atValidElement( pIter );
				PIDBIterator_next( pIter ) )
				{
				DescriptionMap *pDesc = (DescriptionMap *)
					PIDBIterator_current( pIter, 0 );
				assert( pDesc );
				PI_DELETE( pDesc );
				};
			};
		if ( pDescriptionDB )
			{ PIDB_delete( pDescriptionDB ); };
		PIDBIterator_delete( pIter );
		return true;
		};

public:

	/*___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ *

					Handle the request

	 *___ +++++++++++++++++++++++++++++++++++++++++++++++++ ___ */
	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer & /* tB */ ) 
		{
		if ( iPhase!=PH_HANDLE ) { return PIAPI_ERROR; };

		PIDB *pC = tPIHTTP.pConnectionDB;
		PIDB *pQ = tPIHTTP.pRequestDB;
		PIDB *pR = tPIHTTP.pResponseDB;

		/* --- 
		The following is the kernel of the sort property
		together with the qSort function.
		--- */

		// set the default sort order by options value
		bSortDescending = ( iOptions & FLG_SORTDESC );			

		// parse the query string and change the table header row
		char *pQueryString = (char *)
			PIDB_lookup( pQ, PIDBTYPE_STRING, KEY_HTTP_QUERYSTRING, 0 );
		char *pos = NULL;
		int bListTopModified = 0;

		// set the default sort column by options value
		// and find default field in table header row
		if ( iOptions & FLG_SORTTYPE )
			{ 
				iSortByColumn = SORT_TYPE;
				pos = strstr( sListTop, KEY_SORT_TYPE );
			}
		else if ( iOptions & FLG_SORTNAME )
			{
				iSortByColumn = SORT_NAME;
				pos = strstr( sListTop, KEY_SORT_NAME );
			}
		else if ( iOptions & FLG_SORTSIZE )
			{ 
				iSortByColumn = SORT_SIZE;
				pos = strstr( sListTop, KEY_SORT_SIZE );
			}
		else if ( iOptions & FLG_SORTDATE )
			{
				iSortByColumn = SORT_DATE;
				pos = strstr( sListTop, KEY_SORT_DATE );
			}
		else if ( iOptions & FLG_SORTDSCR )
			{
				iSortByColumn = SORT_DESC;
				pos = strstr( sListTop, KEY_SORT_DESC );
			};

		// the last sort order
//		int iOldColumn = iSortByColumn;
		
		// expected Tokens in QueryString
		const char *pColumn = 0;
		const char *pSort = 0;
		
		if ( !pQueryString || !strcmp( pQueryString, "" )) {
		    if (pos) {
			// set the sort order for the default column to the inverse default
			if ( bSortDescending )
				{ strncpy( pos + SORT_KEY_LEN + 1, "A", 1 ); }
			else
				{ strncpy( pos + SORT_KEY_LEN + 1, "D", 1 ); };
			bListTopModified = 1;
		    };
		} else {
			StringTokenizer tTokens( pQueryString, "=" );
			if (tTokens.NumTokens() < 2) {
				HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Invalid QueryString '%s'.", pQueryString );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			} else {
				pColumn = (const char *)tTokens.GetToken( 0 );
				pSort = (const char *)tTokens.GetToken( 1 );
			};

			// switch sort order			
			if ( pSort ) { bSortDescending = !strcmp( pSort, "D" ); };

	    		// find the current sort column (query string must be part of header row)
			pos = pColumn ? strstr( sListTop, pColumn ) : 0;
			if ( pos ) {
			    // set the sort column by query string field
				if ( !strcmp( pColumn, KEY_SORT_TYPE )) { iSortByColumn = SORT_TYPE; };
				if ( !strcmp( pColumn, KEY_SORT_NAME )) { iSortByColumn = SORT_NAME; };				
				if ( !strcmp( pColumn, KEY_SORT_SIZE )) { iSortByColumn = SORT_SIZE; };
				if ( !strcmp( pColumn, KEY_SORT_DATE )) { iSortByColumn = SORT_DATE; };
				if ( !strcmp( pColumn, KEY_SORT_DESC )) { iSortByColumn = SORT_DESC; };
    
				// the header has to be patched
				if ( bSortDescending ) {
				    strncpy( pos + strlen( pColumn ) + 1, "A", 1 );
				} else {
				    strncpy( pos + strlen( pColumn ) + 1, "D", 1 );
		    		};
    				bListTopModified = 1;
			};
		};
		
		// write the changed table header into Pi3-Expression
		if ( bListTopModified ) {
			// delete old expression
			if ( pListTopPattern ) { Pi3Expression_delete( pListTopPattern ); };
			Pi3String *pError = Pi3String_new( 0 );

			// create new expression from changed table header
			pListTopPattern = Pi3Expression_new( sListTop, aFunctions, pError );
			Pi3String_delete( pError );

			// Error!
			if ( !pListTopPattern )	{
				HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Error when trying to change header row." );
				return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );;
			};
		};

		/* --- request must be GET or HEAD --- */
		int iMethod = (int)PIDB_lookup( pQ, PIDBTYPE_OPAQUE, KEY_HTTP_METHOD,
			0 );
		if ( iMethod!=MD_GET && iMethod!=MD_HEAD )
			{ return PIAPI_CONTINUE; };

		/* ---
		Get URL Path
		--- */
		const char *pURI = (const char *)
			PIDB_lookup( pQ, PIDBTYPE_STRING, pFKURI, PIDBFLAG_FASTKEY ); 
		if ( !pURI )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: pURI is NULL." );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );;
			};

		/* --- get directory --- */
		const char *pPath = (const char *)PIDB_lookup( pR, PIDBTYPE_STRING,
			pFKPath, PIDBFLAG_FASTKEY );	
		PIFInfo *pDir = HTTPCore_getCachedFile( pPath );

		if ( !pDir )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: \
HTTPCore_getCachedFile() failed.");
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		if ( !PIFInfo_exists( pDir ) )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Resource \
with path '%s' does not exist.", PIFInfo_getPath( pDir ) );
			HTTPCore_releaseCachedFile( pDir );
			return HTTPUtil_doHTTPError( &tPIHTTP, 404 );
			};

		if ( !PIFInfo_isDirectory( pDir ) )
			{
			HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Resource \
with path '%s' is not a directory.", PIFInfo_getPath( pDir ) );
			HTTPCore_releaseCachedFile( pDir );
			return HTTPUtil_doHTTPError( &tPIHTTP, ST_INTERNALERROR );
			};

		/* --- set the status --- */
		tPIHTTP.iStatus = ST_OK;

		/* --- replace content-type --- */
		PIDB_replace( pR, PIDBTYPE_RFC822, pFKContentType,
			(void *)"text/html", PIDBFLAG_FASTKEY );

		/* --- send last modified --- */
		Pi3String *pTmp = Pi3String_new( 0 );
		time_t tT = PIFInfo_getLastModified( pDir );
		PIPlatform_beforeUnsafeBlock();
		struct tm *pTms = gmtime( &tT );
		HTTPUtil_rFC822Time( pTms, pTmp );
		PIPlatform_afterUnsafeBlock();
		PIDB_add( pR, PIDBTYPE_RFC822, KEY_HTTP_LASTMODIFIED,
			(void *)Pi3String_getPtr( pTmp ), 0 );
		Pi3String_delete( pTmp );

		/* ---
		since the content-length could not be send, the connection
		cannot be kept open 
		--- */
		PIDB_replace( pC, PIDBTYPE_OPAQUE, KEY_INT_KEEPOPEN, (void *)0, 0 );

		/* --- replace headers --- */
        for( DblListIterator i( lHTTPEquivalents ); !i.BadIndex(); i++)
            { ((RFC822Header *)i.Current())->ReplaceHeader( pR ); };

		/* --- transfer-encoding --- */
		int iFlags = (!PIHTTP_isChild( &tPIHTTP) && (int)PIDB_lookup( tPIHTTP.pRequestDB,
			PIDBTYPE_OPAQUE, KEY_HTTP_PROTOCOL, 0 ) == 3 ) ? PIIOBUF_CHUNKED : PIIOBUF_NONE;
		if ( iFlags & PIIOBUF_CHUNKED )
			{
			PIDB_replace( pR, PIDBTYPE_RFC822, KEY_HTTP_TRANSFERENCODING,
				(void *)KEY_HTTP_CHUNKED, 0 );
			};

		/* --- start response --- */
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{
			HTTPCore_releaseCachedFile( pDir );
			return PIAPI_ERROR;
			};

		/* ---
		If this is a GET request then generate the response
		--- */
		if ( iMethod==MD_GET )
			{
			if ( iFlags & PIIOBUF_CHUNKED ) {
				PIIOBuffer_flush( tPIHTTP.pBuffer );
				// we don't write any byte, only switch on chunked mode
				PIIOBuffer_write( tPIHTTP.pBuffer, pURI, 0, PIIOBUF_CHUNKED );
			};

			/* ---
			Generate context 
			--- */
			WriteContext tContext;

			if ( !GenerateDirectoryListing( tPIHTTP, tContext, pDir, pURI ) )
				{

				/* ---
				its too late for a message, write a message to the error
				log instead 
				--- */
				HTTPCore_logError( &tPIHTTP, "DirectoryIndex: Error \
generating directory listing." );
				};
			};

		/* --- 
		Done!
		--- */
		PIIOBuffer_flush( tPIHTTP.pBuffer );
		HTTPCore_releaseCachedFile( pDir );
		return PIAPI_COMPLETED;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int DirectoryIndex_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( DirectoryIndex( pObj,
		iArgc, ppArgv ) ) );
}



#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name DirectoryIndexClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor DirectoryIndex_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name DirectoryIndex
		Class DirectoryIndexClass
	</Object>

/*___+++CNF_END+++___*/
#endif

