/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Platform/KrnlSync.cpp,v $
 * $Date: 2003/06/07 09:05:31 $
 *
 Description:
	This file implements threads using a natively supplied facility. The
	facilities used are Win32 threads, solaris (UNIX SVR4.2 MP) and
	pthreads (POSIX 1003.1c). Although these interfaces have very different
	functions, the semantics are largely the same. So they are handled
	in the same file with lots of #if ... #endif.

\*____________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"		/* --- get build options --- */

/* ---
 Include this file only for builds which use some native multithreading
	technique
--- */
#if defined(CONFIG_MULTITHREADED) && !defined(CONFIG_MT_USER)

#include <assert.h>
#include <errno.h>
#include <time.h>
#include <iostream.h>

#if !defined(CONFIG_CPP_EXCEPTIONS)
#	include <setjmp.h>
#endif

#if defined(CONFIG_MT_PTHREAD)
#	include <pthread.h>
#	include <signal.h>

#elif defined(CONFIG_MT_SOLARIS)
#	include <thread.h>
#	include <synch.h> 
#	include <unistd.h>
#	include <signal.h>
#	if (CONFIG_LOCK_FLOCK)
#		include <sys/file.h>
#	endif
#	include <sys/types.h>
#	include <sys/wait.h>

#elif defined(CONFIG_MT_WIN32)
#	include <windows.h>
#	include <process.h>

#else
#	error Unsupported!
#endif

#include "PICompat.h"
#include "Allocate.h"
#include "OSEntity.h"
#include "Platform.h"
#include "PlatDefs.h"

/*____________________________________________________________________________*\
 *
 Description:
	Unix thread priorities
\*____________________________________________________________________________*/

#if defined(CONFIG_MT_PTHREAD)
#define THREAD_PRIORITY_LOWEST			1
#define THREAD_PRIORITY_BELOW_NORMAL	25
#define THREAD_PRIORITY_MED				50
#define THREAD_PRIORITY_NORMAL			THREAD_PRIORITY_MED
#define THREAD_PRIORITY_ABOVE_NORMAL	75
#define THREAD_PRIORITY_HIGHEST			99

#elif !defined(CONFIG_MT_WIN32)
#define THREAD_PRIORITY_LOWEST			1
#define THREAD_PRIORITY_BELOW_NORMAL	32	
#define THREAD_PRIORITY_MED				48	
#define THREAD_PRIORITY_NORMAL			THREAD_PRIORITY_MED
#define THREAD_PRIORITY_ABOVE_NORMAL	64
#define THREAD_PRIORITY_HIGHEST			127

#elif defined(CONFIG_MT_WIN32)
#endif

/*____________________________________________________________________________*\
 *
 Typedefs:
\*____________________________________________________________________________*/
#if defined(CONFIG_MT_PTHREAD)
	typedef pthread_t THREAD_T;
	typedef pthread_key_t THREAD_KEY_T;
#	define BAD_THREAD ((THREAD_T)-1)

#elif defined(CONFIG_MT_SOLARIS)
	typedef thread_t THREAD_T;
	typedef thread_key_t THREAD_KEY_T;
#	define BAD_THREAD ((THREAD_T)-1)

#elif defined(CONFIG_MT_WIN32)
	typedef HANDLE THREAD_T;
	typedef DWORD THREAD_KEY_T;
#	define BAD_THREAD ((THREAD_T)0)

#else
#	error Unsupported!	
#endif

/*____________________________________________________________________________*\
 *
 Declarations:
\*____________________________________________________________________________*/
static THREAD_KEY_T tTlsAllocatorKey;
static THREAD_KEY_T tThreadObjectKey;
#if !defined(CONFIG_CPP_EXCEPTIONS)
static jmp_buf ___tJmpBuf;
#endif
/*
** #define D { cerr << __FILE__ << ": " << __LINE__ << endl; }
*/
#define D

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline bool Internal_SetData( THREAD_KEY_T tKey, void *pData )
{
#if defined(CONFIG_MT_PTHREAD)
	int iRet = ::pthread_setspecific( tKey, pData );
	if ( iRet )
		{
		/* --- cannot use PIRETERR, because of danger of recursion --- */
		assert( 0 );
		return false;
		}
	else
		{
		return true;
		};

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_setspecific( tKey, pData );
	if ( iRet )
		{
		/* --- cannot use PIRETERR, because of danger of recursion --- */
		assert( 0 );
		return false;
		}
	else
		{
		return true;
		};

#elif defined(CONFIG_MT_WIN32)
	if ( !::TlsSetValue( tKey, pData ) )
		{ PIOSERR; return false; }
	else return true;

#else
#	error Unsupported!	
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
inline bool Internal_GetData( THREAD_KEY_T tKey, void **ppData )
{
#if defined(CONFIG_MT_PTHREAD)
	*ppData = ::pthread_getspecific( tKey );
	if ( !*ppData && errno )
		{
		assert( 0 );
		return false;
		}
	else
		{ return true; };

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_getspecific( tKey, ppData );
	if ( iRet )
		{
		/* --- cannot use PIRETERR, because danger of recursion --- */
		assert( 0 );
		return false;
		}
	else
		{ return true; };

#elif defined(CONFIG_MT_WIN32)
	*ppData = ::TlsGetValue( tKey );
	if ( !*ppData && GetLastError() ) { PIOSERR; };
	return true;

#else
#	error Unsupported!
#endif
}

#if 0
/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	This function will only be called after Platform::Internal_InitThreads so
	it is safe to assume the thread key tTlsAllocator has been setup.
\*____________________________________________________________________________*/
Allocator *Platform::GetCurrentAllocator()
{
	return GetGlobalAllocator();
}
#endif

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class POSIXLocalMutex : public Semaphore
{
private:
#if defined(CONFIG_MT_PTHREAD)
	pthread_mutex_t tMutex;
#elif defined(CONFIG_MT_SOLARIS)
	mutex_t tMutex;
#elif defined(CONFIG_MT_WIN32)
	HANDLE hMutex;
#else
#	error Unsupported!	
#endif
	bool bError;

public:
	POSIXLocalMutex()
	:	bError( false )
		{
#	if defined(CONFIG_MT_PTHREAD)
#if 0
	/*
	** This mutex initialization is a little old
	*/
		pthread_mutexattr_t tAttr;
		tAttr.m_type = MUTEX_TYPE_STATIC_FAST;
		tAttr.m_flags = MUTEX_FLAGS_PRIVATE;
#endif
		int iRet = ::pthread_mutex_init( &tMutex, 0 );
		if ( iRet!= 0 )
			{ PIRETERR(iRet); bError = true; };

#	elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::mutex_init( &tMutex, USYNC_THREAD, 0);
		if ( iRet != 0 )
			{ PIRETERR(iRet); bError = true; };

#	elif defined(CONFIG_MT_WIN32)
		hMutex = ::CreateMutex( 0, FALSE, 0 );
		if ( !hMutex )
			{ PIOSERR; bError = true; };
//		::InitializeCriticalSection( &tCS );

#	else
#		error Unsupported!	
#	endif
		};

	virtual ~POSIXLocalMutex()
		{
#	if defined(CONFIG_MT_PTHREAD)
		int iRet = ::pthread_mutex_destroy( &tMutex );
		assert( !iRet );
		if ( iRet ) { PIRETERR(iRet); };

#	elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::mutex_destroy( &tMutex );
		assert( !iRet );
		if ( iRet ) { PIRETERR(iRet); };

#	elif defined(CONFIG_MT_WIN32)
		BOOL bRet = ::CloseHandle( hMutex );
		assert( bRet );
		if ( !bRet ) { PIOSERR; };
//		DeleteCriticalSection( &tCS );

#	else
#		error Unsupported!	
#	endif
		};

	virtual int Lock()
		{
#	if defined(CONFIG_MT_PTHREAD)
		int iRet = ::pthread_mutex_lock( &tMutex );
		if ( iRet != 0 )	
			{ PIRETERR(iRet); bError = true; };
		return iRet;

#	elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::mutex_lock( &tMutex );
		if ( iRet != 0 )	
			{ PIRETERR(iRet); bError = true; };
		return iRet;

#	elif defined(CONFIG_MT_WIN32)
		if ( ::WaitForSingleObject( hMutex, INFINITE ) == WAIT_FAILED )
			{ PIOSERR; bError = true; };
		return bError ? -1 : 0;

#	else
#		error Unsupported!	
#	endif
		};	

	virtual int TryLock()
		{
#	if defined(CONFIG_MT_PTHREAD)
		int iRet = pthread_mutex_trylock( &tMutex );
		if ( !iRet ) 
			{ return 0; }
		else if ( iRet == EBUSY )
			{ return PISYNC_EBUSY; }
		else
			{ PIRETERR(iRet); bError = true; return -1; };

#	elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::mutex_trylock( &tMutex );
		if ( !iRet ) 
			{ return 0; }
		else if ( iRet == EBUSY )
			{ return PISYNC_EBUSY; }
		else
			{ PIRETERR(iRet); bError = true; return -1; };

#	elif defined(CONFIG_MT_WIN32)
		DWORD dwResult = ::WaitForSingleObject( hMutex, 0 );
		if ( dwResult == WAIT_TIMEOUT )
			{ return PISYNC_EBUSY; }
		else if ( dwResult != WAIT_FAILED )
			{ return 0; }
		else
			{ PIOSERR; bError = true; return -1; };

#	else
#		error Unsupported!	
#	endif
		};	

	virtual int UnLock()
		{
#	if defined(CONFIG_MT_PTHREAD)
		int iRet = ::pthread_mutex_unlock( &tMutex );
		if ( iRet!=0 )
			{ PIRETERR(iRet); bError = true; };
		return iRet;	

#	elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::mutex_unlock( &tMutex );
		if ( iRet!=0 )
			{ PIRETERR(iRet); bError = true; };
		return iRet;	

#	elif defined(CONFIG_MT_WIN32)
		bError = ::ReleaseMutex( hMutex ) != TRUE;
		if ( bError ) { PIOSERR; };
		return bError ? -1 : 0;

#	else
#		error Unsupported!	
#	endif
		};

	virtual bool IsOK() const
		{ 
		return !bError;
		};
};

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
#if !defined(CONFIG_MT_PTHREAD)
class POSIXLocalSemaphore : public Semaphore
{
private:
#if defined(CONFIG_MT_PTHREAD)
#elif defined(CONFIG_MT_SOLARIS)
	sema_t tSemaphore;
#elif defined(CONFIG_MT_WIN32)
	HANDLE hSemaphore;
#else
#	error Unsupported!	
#endif
	bool bError;

public:
	POSIXLocalSemaphore( int iInitialCount, int iMaxCount );
	virtual ~POSIXLocalSemaphore();
	virtual int Lock();
	virtual int TryLock();
	virtual int UnLock();
	virtual bool IsOK() const;
};
#endif

/*____________________________________________________________________________*\
 *
 Declarations:
\*____________________________________________________________________________*/
unsigned long PIPOSIXThreadFunction( unsigned long );

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class POSIXThread : public Thread
{
public:
	class StackUnwindException
		{
		};

private:
	THREAD_T tThread;
	ThreadFn fnThread;
	Allocator *pAllocator;
	unsigned long ulData;
	int iStackSize;
	bool InternalSetPriority( int iPriority );
	bool IsTerminated();

public:
	/* --- some public internal functions --- */
	static bool InternalWaitForJoin( THREAD_T tThread );

	POSIXThread();		/* --- special constructor for main thread --- */
	POSIXThread( Allocator *pAllocator, int iStackSize );
	virtual ~POSIXThread();
	virtual bool Begin( ThreadFn fnEntry, unsigned long ulData, 
		int iPriority = PITHREAD_PRIORITY_MED, int iFlags=0 );
	virtual bool Terminate( unsigned long ulExitCode );
	virtual bool Suspend();
	virtual bool Resume();
	virtual int GetPriority()
		{
		int iPhysicalPriority;
#if defined(CONFIG_MT_PTHREAD)

		sched_param schedParam;
		int iPolicy;

		int iRet = ::pthread_getschedparam( tThread, &iPolicy, &schedParam );
		iPhysicalPriority = schedParam.sched_priority;
		if ( iRet )
			{
			PIOSERR;
			return PITHREAD_PRIORITY_INVALID;
			}

#elif defined(CONFIG_MT_SOLARIS)
		int iRet = ::thr_getprio( tThread, &iPhysicalPriority );
		if ( iRet )
			{
			PIRETERR( iRet );
			return PITHREAD_PRIORITY_INVALID;
			}

#elif defined(CONFIG_MT_WIN32)
		iPhysicalPriority = 
			::GetThreadPriority( tThread );
		if ( iPhysicalPriority == THREAD_PRIORITY_ERROR_RETURN )
			{
			PIOSERR;
			return PITHREAD_PRIORITY_INVALID;
			}

#else
#	error Unsupported!
#endif
		else if ( iPhysicalPriority <= THREAD_PRIORITY_LOWEST )
			return PITHREAD_PRIORITY_LOWEST;
		else if ( iPhysicalPriority <= THREAD_PRIORITY_BELOW_NORMAL )
			return PITHREAD_PRIORITY_LOW;
		else if ( iPhysicalPriority <= THREAD_PRIORITY_NORMAL )
			return PITHREAD_PRIORITY_MED;
		else if ( iPhysicalPriority <= THREAD_PRIORITY_ABOVE_NORMAL )
			return PITHREAD_PRIORITY_HIGH;
		else if ( iPhysicalPriority <= THREAD_PRIORITY_HIGHEST ) 
			return PITHREAD_PRIORITY_HIGHEST;
		else
			{
#if !defined(NDEBUG)
			cerr << "Unknown priority #" << iPhysicalPriority << endl;
			assert( 0 );
#endif
			return PITHREAD_PRIORITY_INVALID;
			};
		};

	virtual bool SetPriority( int iPriority );
	virtual bool SetData( int iKey, void *pData )
		{ return Internal_SetData( iKey, pData ); };
	virtual bool GetData( int iKey, void **ppData )
		{ return Internal_GetData( iKey, ppData ); };
	virtual void *GetSystemHandle()
		{
		return (void *)tThread;
		};
	
	/* --- new functions --- */
	unsigned long ThreadFunction();	
	void Cleanup()
		{
#if defined(CONFIG_MT_SOLARIS) || defined(CONFIG_MT_PTHREAD)
		tThread = BAD_THREAD;

#elif defined(CONFIG_MT_WIN32)
		if ( tThread!=BAD_THREAD )
			{
			if ( !::CloseHandle( tThread ) )
				{
				PIOSERR;
				assert( 0 );
				};
			tThread = BAD_THREAD;
			};

#endif
		};

	/* --- friends --- */
	friend class Platform;
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
static POSIXThread *___pTheMainThread = 0;
void Internal_InitThreads()
{
	int iTmp;
	iTmp = Platform::AllocThreadKey();
	PIMUSTBETRUE( iTmp >=0 );
	tTlsAllocatorKey = iTmp;
	iTmp = Platform::AllocThreadKey();
	PIMUSTBETRUE( iTmp >=0 );
	tThreadObjectKey = iTmp;
	if ( tTlsAllocatorKey==tThreadObjectKey )
		{
		cerr << "##" << endl << "## Error in native threads library." << endl;
		cerr << "##" << endl << "## thr_keycreate() may not be implemented \
correctly." << endl;
		cerr << "##" << endl << "## Please check ld library configuration."
			<< endl;
		cerr << "##" << endl;
		abort();
		};
	assert( tTlsAllocatorKey!=tThreadObjectKey );

	PIMUSTBETRUE(
		Internal_SetData( tTlsAllocatorKey, Platform::GetGlobalAllocator() ) );
	assert( !___pTheMainThread );
	___pTheMainThread = PI_NEW( POSIXThread() );
	PIMUSTBETRUE( Internal_SetData( tThreadObjectKey, ___pTheMainThread ) );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Internal_CleanupThreads()
{
	void *pMainThread = 0;
	PIMUSTBETRUE( Internal_GetData( tThreadObjectKey, &pMainThread ) );
	assert( ___pTheMainThread==pMainThread );
	PIMUSTBETRUE( Platform::DeleteThreadKey( tTlsAllocatorKey ) );
	PIMUSTBETRUE( Platform::DeleteThreadKey( tThreadObjectKey ) );
	PI_DELETE( (POSIXThread *)pMainThread );
	___pTheMainThread = 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
void Platform::Internal_EnterThreads( void (* fn)(void) )
{
	Internal_InitThreads();
	Platform::CatchExceptions( (void(*)(void*))fn, 0 );
	Internal_CleanupThreads();
}

#if !defined(CONFIG_MT_PTHREAD)
/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	public constructor
 Description:
\*____________________________________________________________________________*/
POSIXLocalSemaphore::POSIXLocalSemaphore( int iInitialCount, int iMaxCount )
:	bError( true )
{
#if defined(CONFIG_MT_SOLARIS)
	int iRet = ::sema_init(
		&tSemaphore, 
		iMaxCount,
		USYNC_THREAD,
		0 );
	bError = ( iRet!=0 );
	if ( bError ) { PIRETERR(iRet); return; };
	int iSemaphoresToLock=iMaxCount-iInitialCount;
	assert( iSemaphoresToLock >= 0 );
	while( iSemaphoresToLock-- )
		{ PIMUSTBETRUE(!::sema_trywait( &tSemaphore ));};

#elif defined(CONFIG_MT_WIN32)
	hSemaphore = ::CreateSemaphore( 0, iInitialCount, iMaxCount, 0 );
	bError = !hSemaphore;
	if ( bError ) 
		{ PIOSERR; return; };

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
POSIXLocalSemaphore::~POSIXLocalSemaphore()
{
#if defined(CONFIG_MT_SOLARIS)
	int iRet = ::sema_destroy( &tSemaphore );
	if ( iRet )
		{ PIRETERR(iRet); };

#elif defined(CONFIG_MT_WIN32)
	if ( !::CloseHandle( hSemaphore ) ) { PIOSERR; };

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int POSIXLocalSemaphore::Lock()
{
#if defined(CONFIG_MT_SOLARIS)
	for(;;)
		{
		int iRet = ::sema_wait( &tSemaphore );
		if ( iRet == EINTR )
			{ continue; };
		if ( iRet != 0 )
			{ PIRETERR(iRet); bError = true; };
		return iRet;
		};

#elif defined(CONFIG_MT_WIN32)
	if ( ::WaitForSingleObject( hSemaphore, INFINITE ) == WAIT_FAILED )
		{ PIOSERR; bError = true; };
	return bError ? -1 : 0;

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int POSIXLocalSemaphore::TryLock()
{
#if defined(CONFIG_MT_SOLARIS)
	int iRet = ::sema_trywait( &tSemaphore );
	if ( iRet == EBUSY )
		{ return PISYNC_EBUSY; }
	else if ( iRet == 0 )
		{ return 0; }
	else
		{ PIRETERR(iRet); bError = true; return -1; };

#elif defined(CONFIG_MT_WIN32)
	DWORD dwResult = ::WaitForSingleObject( hSemaphore, 0 );
	if ( dwResult == WAIT_TIMEOUT )
		{ return PISYNC_EBUSY; }
	else if ( dwResult != WAIT_FAILED )
		{ return 0; }
	else
		{ PIOSERR; bError = true; return -1; };

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int POSIXLocalSemaphore::UnLock()
{
#if defined(CONFIG_MT_SOLARIS)
	int iRet = ::sema_post( &tSemaphore );
	if ( iRet )
		{ PIRETERR(iRet); bError = true; };
	return iRet;

#elif defined(CONFIG_MT_WIN32)
	bError = ( ::ReleaseSemaphore( hSemaphore, 1, 0 ) != TRUE );
	if ( bError ) { PIOSERR; };
	return bError ? -1 : 0;

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool POSIXLocalSemaphore::IsOK() const
{
	return !bError;
}
#endif	/* CONFIG_MT_PTHREAD */

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	standalone function
 Description:
	Entry point for threads
\*____________________________________________________________________________*/
unsigned long PIPOSIXThreadFunction( unsigned long ulThread )
{
#if defined(CONFIG_MT_PTHREAD)

#elif defined(CONFIG_MT_SOLARIS)
#if 0
	/*
	** NOTE:
	** This is not right yet
	**
	*/
	/* --- block some signals from this thread --- */
	sigset_t tSet;
	sigemptyset( &tSet );
	thr_sigsetmask( SIG_BLOCK, &tSet, NULL );
#endif

#elif defined(CONFIG_MT_WIN32)
#else
#	error Unsupported!
#endif

	POSIXThread *pThread=(POSIXThread *)ulThread;
	assert( pThread );
	Internal_SetData( tThreadObjectKey, pThread );
	unsigned long ulRet=pThread->ThreadFunction();	
	return ulRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis: 	public, constructor
 Description:
	Constructor.
\*____________________________________________________________________________*/
POSIXThread::POSIXThread( 
		Allocator *pTheAllocator,
		int iTheStackSize )
:
	tThread( BAD_THREAD ),
	fnThread( 0 ),
	pAllocator( pTheAllocator ),
	ulData( 0 ),
	iStackSize( iTheStackSize )
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis: 	public, constructor
 Description:
	Special constructor for main thread.
\*____________________________________________________________________________*/
POSIXThread::POSIXThread()
:
#if defined(CONFIG_MT_PTHREAD)
	tThread( ::pthread_self() )

#elif defined(CONFIG_MT_SOLARIS)
	tThread( ::thr_self() )

#elif defined(CONFIG_MT_WIN32)
	tThread( ::GetCurrentThread() )

#else
#	error Unsupported!
#endif
	, fnThread( 0 ),
	pAllocator( 0 ),
	ulData( 0 ),
	iStackSize( 0 )
{
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::IsTerminated()
{
#if defined(CONFIG_MT_PTHREAD)
	return ::pthread_kill( tThread, 0 )!=0;
	
#elif defined(CONFIG_MT_SOLARIS)
	return ::thr_kill( tThread, 0 )!=0;	/* --- checks if thread exists --- */

#elif defined(CONFIG_MT_WIN32)
	if ( tThread==BAD_THREAD )
		{ return true; };
	DWORD dwExitCode;
	if ( ::GetExitCodeThread( tThread, &dwExitCode ) )
		{
		return dwExitCode!=STILL_ACTIVE;
		}
	else
		{
		PIOSERR;
		assert( 0 );
		return true;
		};

#else
#	error Unsupported!
#endif

}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis: 	public, constructor
 Description:
	Constructor.
\*____________________________________________________________________________*/
bool POSIXThread::Begin(
		ThreadFn fnEntry,
		unsigned long ulTheData,
		int iPriority,
		int iTheFlags )
{
	if ( !IsTerminated() )
		{ return false; };

	if ( !fnEntry )
		{ return false; };

	Cleanup();

	fnThread = fnEntry;
	ulData = ulTheData;

#if defined(CONFIG_MT_PTHREAD)
	pthread_attr_t tAttr;
	::pthread_attr_init( &tAttr );
	(void)iTheFlags;
	int iRet = ::pthread_create(
		&tThread,	
		&tAttr,
		(void*(*)(void*))PIPOSIXThreadFunction, 
		(void *)this );
	::pthread_attr_destroy( &tAttr );
	if ( iRet )
		{
		PIRETERR(iRet);
		return false;
		};

#elif defined(CONFIG_MT_SOLARIS)
	int iFlags = THR_NEW_LWP;	/* for kernel threads */
	if ( iTheFlags & PITHREAD_FLAGS_SUSPENDED )
		{ iFlags |= THR_SUSPENDED; };
	int iRet = ::thr_create(
		0,
		iStackSize, 
		(void*(*)(void*))PIPOSIXThreadFunction, 
		(void *)this,
		iFlags,
		&tThread );
	if ( iRet )
		{ PIRETERR(iRet); return false; };

#elif defined(CONFIG_MT_WIN32)
	int iFlags = 0;
	if ( iTheFlags & PITHREAD_FLAGS_SUSPENDED )
		{ iFlags |= CREATE_SUSPENDED; };
	unsigned int uiThreadId;
	if ( (tThread = (THREAD_T)::_beginthreadex(
					NULL,				/* no security attributes */
					iStackSize,			/* default stack size */
					(unsigned int (__stdcall *)(void *))PIPOSIXThreadFunction,
					(LPVOID)this,		/* pass this pointer to ThreadFn */
					iFlags,				/* creation flags */
					&uiThreadId ) )==BAD_THREAD )
		{ PIOSERR; return false; };

#else
#	error Unsupported!
#endif

	/* --- set the priority --- */
	if ( !InternalSetPriority( iPriority ) )
		{
#if defined(CONFIG_MT_PTHREAD)
		::pthread_kill( tThread, SIGTERM );

#elif defined(CONFIG_MT_SOLARIS)
		::thr_kill( tThread, SIGTERM );

#elif defined(CONFIG_MT_WIN32)
		::TerminateThread( tThread, 0 );
		Cleanup();

#else
#	error Unsupported!
#endif
		return false;
		};

	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
POSIXThread::~POSIXThread()
{
	if ( ___pTheMainThread!=this )
		{
		if ( !IsTerminated() )
			{
			Terminate( 0 );
			InternalWaitForJoin( tThread );
			};
		Cleanup();
		assert( tThread==BAD_THREAD );
		};
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public
 Description:
\*____________________________________________________________________________*/
unsigned long POSIXThread::ThreadFunction()
{
	assert( fnThread );
	PIMUSTBETRUE( Internal_SetData( tTlsAllocatorKey, pAllocator ) );
	unsigned long ulRet=(*fnThread)( ulData );
	Cleanup();
	return ulRet;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, private
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::InternalWaitForJoin( THREAD_T tThread )
{
	if ( tThread==BAD_THREAD )
		{ return true; /* --- thread was never created or has finished --- */ };

	void *pV;

#if defined(CONFIG_MT_PTHREAD)
	int iRet = ::pthread_join( tThread, &pV );

	/* ---
	There are two error conditions
	--- */
	if ( !iRet )	{ return true; };

	if ( tThread==BAD_THREAD )
		{ /* thread terminated just before thr_join() */ return true; };

	/* --- error --- */
	PIRETERR(iRet);
	return false;

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_join( tThread, &tThread, &pV );

	/* ---
	There are two error conditions
	--- */
	if ( !iRet )	{ return true; };

	if ( tThread==BAD_THREAD )
		{ /* thread terminated just before thr_join() */ return true; };

	/* --- error --- */
	PIRETERR(iRet);
	return false;

#elif defined(CONFIG_MT_WIN32)

	if ( ::WaitForSingleObject( tThread, INFINITE ) != WAIT_FAILED )
		{ return true; };

	if ( tThread==BAD_THREAD )
		{ /* thread terminated just before join */ return true; };

	/* --- this is an error --- */
	PIOSERR;
	return false;

#else
#	error Unsupported
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::Terminate( unsigned long ulExitCode )
{
	if ( tThread==BAD_THREAD )
		{ return true;	/* already terminated */ };

	if ( ___pTheMainThread == this )
		{
#if defined(CONFIG_CPP_EXCEPTIONS)
		throw PI_NEW( StackUnwindException() );
#else
		longjmp( ___tJmpBuf, 37 );
#endif
		};

#if defined(CONFIG_MT_PTHREAD)
	if ( tThread == ::pthread_self() )
		{ ::pthread_exit( (void *)ulExitCode ); };

	int iRet = ::pthread_kill( tThread, SIGTERM );
	if ( iRet )
		{ tThread=BAD_THREAD; }
	else
		{ PIRETERR(iRet); return false; };

#elif defined(CONFIG_MT_SOLARIS)
	if ( tThread == ::thr_self() )
		{ ::thr_exit( (void *)ulExitCode ); };

	int iRet = ::thr_kill( tThread, SIGTERM );
	if ( iRet )
		{ tThread=BAD_THREAD; }
	else
		{ PIRETERR(iRet); return false; };

#elif defined(CONFIG_MT_WIN32)
	if ( !::TerminateThread( tThread, 0 ) )
		{ PIOSERR; return false; };

#else
#	error Unsupported
#endif
	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::Suspend()
{
#if defined(CONFIG_MT_PTHREAD)
	PIRETERR(PIAPI_NOTSUPPORTED);
	return false;

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_suspend( tThread );
	if ( iRet )
		{
		PIRETERR(iRet);
		return false;
		}
#elif defined(CONFIG_MT_WIN32)
	if ( ::SuspendThread( tThread ) == 0xFFFFFFFF )
		{ PIOSERR; return false; };

#else
#	error Unsupported!
#endif

	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		public 
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::Resume()
{
#if defined(CONFIG_MT_PTHREAD)
	PIRETERR(PIAPI_NOTSUPPORTED);
	return false;

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_continue( tThread );
	if ( iRet )
		{
		PIRETERR(iRet);
		return false;
		}

#elif defined(CONFIG_MT_WIN32)
	if ( ::ResumeThread( tThread ) == 0xFFFFFFFF )
		{ PIOSERR; return false; };

#else
#	error Unsupported!
#endif

	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:		private
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::InternalSetPriority( int iPriority )
{
	int iPOSIXThreadPriority=THREAD_PRIORITY_NORMAL;
	switch( iPriority )
		{
		case PITHREAD_PRIORITY_LOWEST:	
					iPOSIXThreadPriority=THREAD_PRIORITY_LOWEST; break;
		case PITHREAD_PRIORITY_LOW:	
					iPOSIXThreadPriority=THREAD_PRIORITY_BELOW_NORMAL; break;
		case PITHREAD_PRIORITY_HIGH:
					iPOSIXThreadPriority=THREAD_PRIORITY_ABOVE_NORMAL; break;
		case PITHREAD_PRIORITY_HIGHEST:
					iPOSIXThreadPriority=THREAD_PRIORITY_HIGHEST; break;
		case PITHREAD_PRIORITY_MED:
		default:;
		};

#if defined(CONFIG_MT_PTHREAD)
	sched_param schedParam;
	int iPolicy;

	int iRet = ::pthread_getschedparam( tThread, &iPolicy, &schedParam );
	if (!iRet)
		{
		iPolicy = SCHED_FIFO;
		schedParam.sched_priority = iPOSIXThreadPriority;
		iRet = ::pthread_setschedparam( tThread, iPolicy, &schedParam );
		if (!iRet)
			{ return true; }
		else
			{
			if ( EPERM == errno )
				/*
				** Missing privileges to change scheduler params
				*/
				{ return true; }
			else
				{
				/*
				** Another error occured
				*/
				PIOSERR;
				return false;
				}
			}
		}
	else
		{
		PIOSERR;
		return false;
		}

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_setprio( tThread, iPOSIXThreadPriority );
	if ( iRet )
		{
		PIRETERR(iRet);
		return false;
		};

#elif defined(CONFIG_MT_WIN32)
	if ( !::SetThreadPriority( tThread, iPOSIXThreadPriority ) )
		{
		PIOSERR;
		return false;
		};

#endif
	return true;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
bool POSIXThread::SetPriority( int iPriority )
{
	return InternalSetPriority( iPriority );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
Semaphore *Platform::AllocLocalMutex()
{
	return PI_NEW( POSIXLocalMutex() );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
Semaphore *Platform::AllocLocalSemaphore( int iInitialCount, int iMaxCount )
{
#if defined(CONFIG_MT_PTHREAD)
	return AllocGlobalSemaphore( iInitialCount, iMaxCount );
#else
	return PI_NEW( POSIXLocalSemaphore( iInitialCount, iMaxCount ) );
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
Thread *Platform::AllocThread( void *pAllocator, int iStackSize )
{
	return PI_NEW( POSIXThread( (Allocator *)pAllocator, iStackSize ) );
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
void Platform::YieldThread()
{
#if defined(CONFIG_MT_PTHREAD)
	/*
	** NOTE: how to make a pthread yield???
	*/
#	if !defined(CONFIG_NO_YIELD)
		::yield();
#	else
		::sched_yield();
#	endif

#elif defined(CONFIG_MT_SOLARIS)
	::thr_yield();

#elif defined(CONFIG_MT_WIN32)
	::Sleep( 0 );	/* --- This is how Win32 gives up timeslice --- */

#else
#	error Unsupported!
#endif
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
int Platform::WaitForThread( Thread *pThread )
{
	return 
	POSIXThread::InternalWaitForJoin( (THREAD_T)pThread->GetSystemHandle() ) ?
		0 : -1;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
Thread *Platform::GetCurrentThread()
{
	void *pV = 0;
	if ( Internal_GetData( tThreadObjectKey, &pV ) )
		{
		return (Thread *)pV;
		};
	PIERROR( PIAPI_EINVAL );
	return 0;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
Thread *Platform::GetMainThread()
{
	return ___pTheMainThread;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
static void fnDummy( void * ) {};
int Platform::AllocThreadKey()
{
	THREAD_KEY_T tKey = 0;
#if defined(CONFIG_MT_PTHREAD)
	int iRet = ::pthread_key_create( &tKey, fnDummy );
	if ( iRet )
		{
		PIRETERR(iRet);
		assert( 0 );
		return -1;
		}

#elif defined(CONFIG_MT_SOLARIS)
	int iRet = ::thr_keycreate( &tKey, fnDummy );
	if ( iRet )
		{
		PIRETERR(iRet);
		assert( 0 );
		return -1;
		}

#elif defined(CONFIG_MT_WIN32)
	if ( ( tKey = ::TlsAlloc() )==0xFFFFFFFF )
		{ PIOSERR; return -1; }; 

#else
#	error Unsupported!
#endif
	return tKey;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:	  static, public:
 Description:
\*____________________________________________________________________________*/
bool Platform::DeleteThreadKey( int iKey )
{
#if defined(CONFIG_MT_PTHREAD)
	int iRet = ::pthread_key_delete( iKey );
	if ( iRet )	
		{
		PIRETERR(iRet);
		return false;
		};

#elif defined(CONFIG_MT_SOLARIS)
#	if !defined(CONFIG_NO_THR_DELETE)
		int iRet = ::thr_keydelete( iKey );
		if ( iRet )
			{
			PIRETERR(iRet);
			return false;
			};

#	else
		(void)iKey;		/* --- suppress 'arg. not used' warnings --- */
#	endif

#elif defined(CONFIG_MT_WIN32)
	if ( !::TlsFree( iKey ) )
		{ PIOSERR; return false; };

#else
#	error Unsupported!
#endif
	return true;
}

#endif /* defined(CONFIG_MULTITHREADED) && !defined(CONFIG_MT_USER) */


