/*___________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *___________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Platform/Platform.cpp,v $
 * $Date: 2003/05/13 18:42:14 $
 *
 Description:
	This file provides the implementation of the abstract platform interface
	for a specific operating system. Most methods are static.
	
	Custom Memory Management
	========================
	To enable/disable custom memory management, search to the line
	with pattern 'Custom', set value of #if to 0 to disable, 1 to enable

\*___________________________________________________________________________*/
//$SourceTop:$

#include "OSConfig.h"

/*
** System includes
*/
#if defined(CONFIG_OS_POSIX)
#	include <sys/types.h>
#	include <sys/stat.h>
#	include <fcntl.h>
#	include "POSIXOSE.h"

#elif defined(CONFIG_OS_WIN32)
#	include <windows.h>
#	include "Win32OSE.h"

#elif defined(CONFIG_OS_WIN16)
#	include <windows.h>
#	include <io.h>
#	include <process.h>
#	include <dir.h>
#	include "Win16OSE.h"
#	define PATH_MAX MAXPATH

#endif

/*
** C library includes
*/
#include <ctype.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

/*
** Local includes
*/
#include "PICompat.h"
#include "Platform.h"
#include "PlatDefs.h"
#include "PIStrStr.h"
#if defined(CONFIG_POSIXISH)
#	include "POSIXFSE.h"
#else
#	include "Win32FSE.h"
#endif

/* --- static data members --- */
static Allocator *___pTheGlobalAllocator = 0;	/* global allocator */
static FileSysType *pFSType=0;			/* derived file system type for OS */
static Semaphore *pUnsafeBlockMutex=0;	/* guard unsafe system/library calls */
#if defined(CONFIG_MULTITHREADED)
static int ___tLastErrorKey;			/* for thread specific last error */
#else
static int ___iLastError;		   		/* for last error code */
#endif
static time_t tLastTime = 0;
static int iLastCount = 0;

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
	Most methods are static, so the constructor has little to do.
\*___________________________________________________________________________*/
Platform::Platform()
{
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, private
 Description:
	Sets up the platform description string.
\*___________________________________________________________________________*/
static PIString *pInternalPlatformDescription=0;
static void Internal_InitDescription()
{
	PIOStrStream os;
	os << PLATFORM_OS << " (" << PLATFORM_MACHINE << ")" << ends;
	pInternalPlatformDescription = PI_NEW( PIString( os.str() ) );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, private
 Description:
	Cleans up the platform description string.
\*___________________________________________________________________________*/
static void Internal_CleanupDescription()
{
	PI_DELETE( pInternalPlatformDescription );
	pInternalPlatformDescription = 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
	Returns a string which is some concatenation of the operating system
	and hardware machine names.
\*___________________________________________________________________________*/
const char *Platform::GetDescription()
{
	assert( pInternalPlatformDescription );
	if ( !pInternalPlatformDescription ) 
		{ return 0; };
	return *pInternalPlatformDescription;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return 0 on success, else error code.
\*___________________________________________________________________________*/
int Platform::BeforeUnsafeBlock()
{
	assert( pUnsafeBlockMutex );
	assert( pUnsafeBlockMutex->IsOK() );
	return pUnsafeBlockMutex->Lock();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Return 0 on success, else error code.
\*___________________________________________________________________________*/
int Platform::AfterUnsafeBlock()
{
	assert( pUnsafeBlockMutex );
	assert( pUnsafeBlockMutex->IsOK() );
	return pUnsafeBlockMutex->UnLock();
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
bool Platform::GetTimeString( PIString &sResult)
{
	time_t timer;
	if ( time( &timer ))
		{
#if defined(CONFIG_MULTITHREADED) && !defined(CONFIG_MT_USER)
		PIMUSTBETRUE( !BeforeUnsafeBlock() );
#endif
		sResult = ctime( &timer );
#if defined(CONFIG_MULTITHREADED) && !defined(CONFIG_MT_USER)
		PIMUSTBETRUE( !AfterUnsafeBlock() );
#endif
		return true;
		}
	else
		{
		return false;
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
bool Platform::GetPrettyTimeString( PIString &sResult )
{
	time_t t;
	struct tm *local;
	enum { LEN_DATA=1023 };
	char szDataBuf[LEN_DATA+1];

	if ( time( &t ))
		{
		local = localtime(&t);
		strftime(szDataBuf, LEN_DATA,
			// Mon Nov 29 01:35 GMT
			"%a %b %d %H:%M %Z",
			local);
		szDataBuf[LEN_DATA]='\0';
		sResult = szDataBuf;
		return true;
		}
	else
		{
		return false;
		};
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Allocate memory. This function is used to interface operating system
	functionality for allocating large chunks of memory. 
\*___________________________________________________________________________*/
void *Platform::AllocMemory( size_t tSize )
{
	return malloc( tSize );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
	Free memory allocated using AllocMemory.
\*___________________________________________________________________________*/
bool Platform::FreeMemory( void *pV )
{
	free( pV );
	return true;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis: 	static, public
 Description:
\*___________________________________________________________________________*/
Allocator *Platform::GetGlobalAllocator()
{
	return ___pTheGlobalAllocator;
}

/*___________________________________________________________________________*\
 *
 Description:
	The following hook functions can be re-assigned to implement
	custom memory management and debug logging.
\*___________________________________________________________________________*/
/* --- these are the hook functions to re-assign --- */
extern void *(* hook_new)( size_t tSize );
extern void (* hook_delete)( void * );
#if !defined(NDEBUG)
extern void (* hook_PIBeforeNew)(int, const char*);
extern void (* hook_PIBeforeDelete)(int, const char*);
extern void (* hook_PIMark)(int, const char*);
#endif

/* --- save the old values in these variables --- */
static void *(* save_hook_new)( size_t tSize )=0;
static void (* save_hook_delete)( void * )=0;
#if !defined(NDEBUG)
static void (* save_hook_PIBeforeNew)(int, const char*)=0;
static void (* save_hook_PIBeforeDelete)(int, const char*)=0;
static void (* save_hook_PIMark)(int, const char*)=0;
#endif

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:	Internal
\*___________________________________________________________________________*/
void (* fnPlatform)(void) = 0;
void Internal_DoPlatform()
{
	pUnsafeBlockMutex = Platform::AllocLocalMutex();
	assert( pUnsafeBlockMutex );
	assert( pUnsafeBlockMutex->IsOK() );
#if defined(CONFIG_OS_POSIX)
	pFSType = PI_NEW( POSIXFileSysType );
#elif defined(CONFIG_OS_WIN32)
	pFSType = PI_NEW( Win32FATFileSysType );
#else /* CONFIG_OS_WIN16) */
	pFSType = PI_NEW( Win16FATFileSysType );
#endif
	Internal_InitDescription();
	assert( fnPlatform );

	/* --- Setup last error functionality --- */
#if defined(CONFIG_MULTITHREADED)
	___tLastErrorKey = Platform::AllocThreadKey();
#endif
	if ( fnPlatform )
		{ (fnPlatform)(); };
#if defined(CONFIG_MULTITHREADED)
	Platform::DeleteThreadKey( ___tLastErrorKey );
#endif

	Internal_CleanupDescription();
	PI_DELETE( pFSType );
	PI_DELETE( pUnsafeBlockMutex );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*___________________________________________________________________________*/
const char *pProgramArgv0 = 0;	/* Used by GNU DLD */
void Platform::Enter( const char *pArgv0, void (* fn)(void) )
{
	/* --- allocate the global allocator --- */
	assert( !___pTheGlobalAllocator );
	___pTheGlobalAllocator = PI_NEW( Allocator() );

	/* --- set the hook functions --- */
	save_hook_new = hook_new;
	save_hook_delete = hook_delete;
#if !defined(NDEBUG)
	save_hook_PIBeforeNew = hook_PIBeforeNew;
	save_hook_PIBeforeDelete = hook_PIBeforeDelete;
	save_hook_PIMark = hook_PIMark;
#endif
	
#if 0	/* Do Custom memory managment ?, 1=yes, 0=no */
	hook_new = PICustomNew;
	hook_delete = PICustomDelete;
#if !defined(NDEBUG)
	hook_PIBeforeNew = PICustomBeforeNew;
	hook_PIBeforeDelete = PICustomBeforeDelete;
	hook_PIMark = PICustomMark;
#endif
#endif

	pProgramArgv0 = pArgv0;
	fnPlatform = fn;
	Internal_EnterThreads( Internal_DoPlatform );

	/* --- reset the hook functions --- */
	hook_new = save_hook_new;
	hook_delete = save_hook_delete;
#if !defined(NDEBUG)
	hook_PIBeforeNew = save_hook_PIBeforeNew;
	hook_PIBeforeDelete = save_hook_PIBeforeDelete;
	hook_PIMark = save_hook_PIMark;
#endif

	/* --- de-allocate the global allocator --- */
	assert( ___pTheGlobalAllocator );
	PI_DELETE( ___pTheGlobalAllocator );
	___pTheGlobalAllocator = 0;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
void *Platform::PICustomNew( size_t tS )
{
	assert( Platform::GetCurrentAllocator() );
	return Platform::GetCurrentAllocator()->Allocate( tS );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::PICustomDelete( void *pData )
{
//  if ( !Platform::GetCurrentAllocator() ) { free( pData ); return; };   // L
#if defined(NDEBUG)
    /* ---
    This for in case of delete 0 which is valid and needs to be handled.
    Don't return for development builds because we still want to write
    something to the logfiles
    --- */
    if ( !pData ) return;
#endif
    assert( GetCurrentAllocator() );
	GetCurrentAllocator()->DeAllocate( pData );
}

#if !defined(NDEBUG)
/*
** These functions are used in development builds only
*/
/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::PICustomBeforeNew( int iLine, const char *pFileName )
{
    if ( !GetCurrentAllocator() )
		{ return; };
    assert( GetCurrentAllocator() );
    GetCurrentAllocator()->BeforeAllocate( iLine, pFileName );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::PICustomBeforeDelete( int iLine, const char *pFileName )
{
    if ( !GetCurrentAllocator() )
		{ return; };
	 assert( GetCurrentAllocator() );
    GetCurrentAllocator()->BeforeDeAllocate( iLine, pFileName );
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::PICustomMark( int iLine, const char *pFileName )
{
    if ( !GetCurrentAllocator() )
		{ return; };
    assert( GetCurrentAllocator() );
    GetCurrentAllocator()->Mark( iLine, pFileName );
}
#endif

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
DLL *Platform::AllocDLL( const char *pPath )
{ 
	assert( pPath );
#if defined(CONFIG_OS_POSIX)
	return PI_NEW( UnixDLL( pPath ) );
#elif defined(CONFIG_OS_WIN32)
	return PI_NEW( Win32DLL( pPath ) );
#else
	return 0;
#endif
}
void *Platform::AllocDLL( void *pPath ) 	// MSVC4.0 hack
{ return Platform::AllocDLL( (const char *)pPath ); }

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:      static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::UserYieldThread()
{
#if defined(CONFIG_MULTITHREADED) && defined(CONFIG_MT_USER)
	Platform::YieldThread();
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:      static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::GetLastError()
{
#if defined(CONFIG_MULTITHREADED)
	int iValue;
	return GetMainThread()->GetData( ___tLastErrorKey, (void **)&iValue ) ?
		iValue : 0;
#else
	return ___iLastError;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:      static, public:
 Description:
\*___________________________________________________________________________*/
void Platform::SetLastError( int iErrorCode )
{
#if defined(CONFIG_MULTITHREADED)
	GetMainThread()->SetData( ___tLastErrorKey, (void *)iErrorCode );
#else
	___iLastError = iErrorCode;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
FileSystemEntity *Platform::GetFile( const PIString &sPath )
{
	assert( pFSType );
	FileSystemEntity *pFSE = 0;
#if defined(CONFIG_POSIXISH)
	pFSE = PI_NEW( POSIXFileSystemEntity( *pFSType, sPath ) );
#elif defined(CONFIG_OS_WIN32)
	pFSE = PI_NEW( Win32FileSystemEntity( *pFSType, sPath ) );
#else
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
#endif
	if ( !pFSE )
		{ PIERROR( PIAPI_EXHAUSTED ); }
	else
		{ PICLEAR; };
	return pFSE;
}
void *Platform::GetFile( void *pV )	/* hack for MSVC4.0, link */
{ return Platform::GetFile( *((const PIString *)pV) ); };

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
MappedFile *Platform::AllocMappedFile( FileSystemEntity &tFSE )
{
#if defined(CONFIG_OS_POSIX)
	return PI_NEW( POSIXMappedFile( tFSE ) );
#elif defined(CONFIG_OS_WIN32)
	return PI_NEW( Win32MappedFile( tFSE ) );
#else
	(void)tFSE;
	PIERROR( PIAPI_NOTSUPPORTED );
	return 0;
#endif
}
void *Platform::AllocMappedFile( char &c )
{ return Platform::AllocMappedFile( (FileSystemEntity &)c ); };

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
	Fork a new process, as per POSIX.
\*___________________________________________________________________________*/
unsigned long Platform::Fork()
{
#if defined(CONFIG_OS_POSIX)
	int i = fork();
	if ( i==-1 ) { return 1; };
	return i;
#else
	return -1;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
unsigned long Platform::GetPID()
{
#if defined(CONFIG_OS_POSIX)
	return (unsigned long)::getpid();
#elif defined(CONFIG_OS_WIN32)
	return (unsigned long)::GetCurrentProcessId();
#else
	return (unsigned long)-1;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::GetUniqueId( const char *szBuf, unsigned int iLen )
{
	enum { BUF_SIZE=64 };
	if (iLen < BUF_SIZE) return BUF_SIZE;
	if (!szBuf) return PIAPI_EINVAL;

	/* ---
	Generates a new id that is a sequence of hexadecimal digits guaranteed
	to be unique for the lifetime of this object.
	--- */
#if defined(CONFIG_MULTITHREADED)
	PIMUSTBETRUE( !BeforeUnsafeBlock() );
#endif
	time_t tDummy;
	time_t tCurrentTime = time( &tDummy );

	if ( tCurrentTime==tLastTime )
		iLastCount++;
	else
		tLastTime=tCurrentTime;
		
	sprintf(
		(char *)szBuf,
		"%lx%lx",
		(long)(tCurrentTime & 0x00FFFFFF),
		(long)iLastCount);
#if defined(CONFIG_MULTITHREADED)
		PIMUSTBETRUE( !AfterUnsafeBlock() );
#endif
	return PIAPI_COMPLETED;
}

#if 0
/*__________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::GetErrno()
{
	return errno;
}
#endif

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
	This function is called early on in the process initialization, it
	perform some actions to make the process be a daemon and to avoid
	unwanted interactions from happening.
\*___________________________________________________________________________*/
bool Platform::DaemonInit()
{
#if defined(CONFIG_OS_POSIX)
	/*
	** Call fork() and then have the parent exit(), this makes us
	** independant of any shell that spawned us, it also makes sure the
	** process is not a process group leader
	*/
	switch( fork() )
		{
		case -1:
			/* error, exit */
			return false;
		case 0:
			/* the child, continue */
			break;
		default:
			/* the parent, exit */
			exit(0);
		};

	/*
	** Do the following actions
	** 	1) Become a session leader
	** 	2) Change the working directory
	** 	3) Clear out the file mode creation mask
	*/
//	setsid( 0 );	?? this doesn't seem to work so well
	setsid( );
	chdir( "/" );
	umask( 0 );

	return true;
#else
	return false;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
	Allocate a timer.
\*___________________________________________________________________________*/
Timer *Platform::AllocTimer()
{
#if defined(CONFIG_OS_POSIX)
	return PI_NEW( UnixTimer() );
#elif defined(CONFIG_OS_WIN32)
	return PI_NEW( Win32Timer() );
#else
	return 0;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Open( const char *pPath, const char *pFlags )
{
	if ( !pPath ) 
		{
		assert( 0 );
		PIERROR( PIAPI_EINVAL );
		return PIAPI_ERROR;
		};
	if ( !pFlags )
		{ pFlags = "w"; };

#if defined(CONFIG_OS_POSIX)
	int iFlags = 0;
#elif defined(CONFIG_OS_WIN32)
	bool bRead = false;
	DWORD dwFlags = 0;
#endif
	for(; *pFlags; pFlags++ )
		{
		switch( toupper( *pFlags ) )
			{
#if defined(CONFIG_OS_POSIX)
			case 'A': iFlags |= O_CREAT | O_WRONLY | O_APPEND; break;
			case 'R': iFlags |= O_RDONLY; break;
			case 'W': iFlags |= O_CREAT | O_WRONLY | O_TRUNC; break;
			case 'N': iFlags |= O_EXCL; break;
#elif defined(CONFIG_OS_WIN32)
			/* --- note that these flags are not OR'ed --- */
			case 'A': dwFlags = OPEN_ALWAYS; break;
			case 'R': dwFlags = OPEN_EXISTING; bRead = true; break;
			case 'W': dwFlags = CREATE_ALWAYS; break;
			case 'N': dwFlags = CREATE_NEW; break;
#endif
			default:
				/* --- invalid flag --- */
				assert( 0 );
				PIERROR( PIAPI_EINVAL );
				return PIAPI_ERROR;
			};
		};
#if defined(CONFIG_OS_POSIX)
	int iFd = -1;
	for(;;)
		{
		iFd = ::open( pPath, iFlags, 0644  );	
		if ( iFd==-1 && errno==EINTR )
			{ continue; };
		break;
		};
	if ( iFd==-1 )
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return iFd;
#elif defined(CONFIG_OS_WIN32)
	HANDLE hFile = INVALID_HANDLE_VALUE;
	if ( !bRead )
		{
		hFile = ::CreateFile(
			pPath,
			GENERIC_WRITE,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			dwFlags,
			FILE_ATTRIBUTE_NORMAL,
			NULL );
		}	
	else
		{
		hFile = ::CreateFile(
			pPath,
			GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			dwFlags,
			FILE_ATTRIBUTE_NORMAL,
			NULL );
		};
	if ( hFile==INVALID_HANDLE_VALUE )
		{
		PIOSERR;
		return PIAPI_ERROR;
		}
	else
		{
		if ( !bRead && ( dwFlags & OPEN_ALWAYS ) )
			{
			/*
			** Append mode was specified, move file pointer to the
			** end of the file
			*/
			DWORD dwWord = ::SetFilePointer( hFile, 0, NULL, FILE_END );
			assert( dwWord>=0 );
			if ( dwWord<0 )
				{
				PIOSERR;
				::CloseHandle( hFile );
				return PIAPI_ERROR;
				};
			};
		};

	return (int)hFile;
	
#endif
	PIERROR( PIAPI_NOTSUPPORTED );
	return PIAPI_ERROR;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Close( int iFd )
{
#if defined(CONFIG_POSIXISH)
	int iRet = -1;
	for(;;)
		{
		iRet = ::close( iFd );
		/*
		** Restart on interupt for UNIX System V
		*/
		if ( iRet==-1 && errno==EINTR )
			{ continue; };
		break;
		};
	if ( iRet==-1 )
#elif defined(CONFIG_OS_WIN32)
	if ( !::CloseHandle( (HANDLE)iFd ) )
#endif
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::Unlink( const char *pPath )
{
	if ( !pPath ) 
		{
		assert( 0 );
		PIERROR( PIAPI_EINVAL );
		return PIAPI_ERROR;
		};

#if defined(CONFIG_POSIXISH)
	int iRet = -1;
	for(;;)
		{
		iRet = ::unlink( pPath );
		/*
		** Restart on interupt for UNIX System V
		*/
		if ( iRet==-1 && errno==EINTR )
			{ continue; };
		break;
		};
	if ( iRet==-1 )
#elif defined(CONFIG_OS_WIN32)
	if ( !::DeleteFile( pPath ) )
#endif
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
	return PIAPI_COMPLETED;
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::RetrieveCurrentDirectory( class PIString &sCurrentDirectory )
{
#if defined(CONFIG_POSIXISH)
	char szBuf[PATH_MAX+1];
	if ( ::getcwd( szBuf, PATH_MAX ) )
		{
		szBuf[PATH_MAX] = '\0';
		sCurrentDirectory = szBuf;
		}
	else
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
#elif defined(CONFIG_OS_WIN32)
	char szBuf[MAX_PATH+1];
	int iLen=::GetCurrentDirectory( MAX_PATH, szBuf );
	if ( iLen>0 )
		{
		szBuf[MAX_PATH] = '\0';
		sCurrentDirectory = szBuf;
		}
	else
		{
		PIOSERR;
		return PIAPI_ERROR;
		};
#endif
	return PIAPI_COMPLETED;
}

char Platform::getDirectorySeparator()
{
	return pFSType ? pFSType->DirectorySeperatorCharacter() : '0';                            
};


/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::GetProcess()
{
#if defined(CONFIG_OS_POSIX)
	return (int)::getpid();
#elif defined(CONFIG_OS_WIN32)
	return (int)::GetCurrentProcess();
#else
	return (int)-1;
#endif
}

/*___________________________________________________________________________*\
 *
 Function:
 Synopsis:		static, public:
 Description:
\*___________________________________________________________________________*/
int Platform::GetProcessId()
{
#if defined(CONFIG_OS_POSIX)
	return (int)::getpid();
#elif defined(CONFIG_OS_WIN32)
	return (int)::GetCurrentProcessId();
#else
	return (int)-1;
#endif
}


