/*____________________________________________________________________________*\
 *

 Copyright (c) 1997-2003 John Roy, Holger Zimmermann. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission. 

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: /cvsroot/pi3web/Pi3Web_200/Source/Servlet/org/pi3/servlet/core/Pi3RequestDispatcher.java,v $
 * $Date: 2003/05/13 18:42:21 $
 *
 Description:
	Implementation of servlet API RequestDispatcher interface.

\*____________________________________________________________________________*/

package org.pi3.servlet.core;

import java.io.IOException;
import javax.servlet.RequestDispatcher;
import javax.servlet.SingleThreadModel;
import javax.servlet.Servlet;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletException;



/**
 * Defines an object that receives requests from the client
 * and sends them to any resource (such as a servlet, 
 * HTML file, or JSP file) on the server. The servlet
 * container creates the <code>Pi3RequestDispatcher</code> object,
 * which is used as a wrapper around a server resource located
 * at a particular path or given by a particular name.
 *
 * <p>The <code>RequestDispatcher</code> interface is intended to wrap
 * servlets, but a servlet container can create <code>RequestDispatcher
 * </code> objects to wrap any type of resource.
 *
 * @author 	Holger Zimmermann [zimpel@pi3.org]
 * @version	$Revision: 1.9 $
 *
 * @see 	Pi3ServletContext#getRequestDispatcher(java.lang.String)
 * @see 	Pi3HttpServletRequest#getRequestDispatcher(java.lang.String)
 *
 */

final public class Pi3RequestDispatcher implements RequestDispatcher {

	private ServletRequest req = null;
	private String url;




    /** The constructor to be called by Pi3ServletContext */
    protected Pi3RequestDispatcher(String urlpath) {
		url = urlpath;
	}




    /** The constructor to be called by Pi3HttpServletRequest */
    protected Pi3RequestDispatcher(ServletRequest request, String path) {
		req = request;
		url = path;
	}




    /** return a Servlet object for a given ressource URL */
    private native Servlet getServlet(String urlpath, ServletRequest request);




    /** call the Webserver to handle a given ressource URL */
    private native void doForward(String urlpath, ServletRequest request,
	    ServletResponse response);



   /**
    * Forwards a request from a servlet to another resource (servlet,
	* JSP file, or HTML file) on the server. This method allows one
	* servlet to do preliminary processing of a request and another
	* resource to generate the response.
    *
    * <p>For a <code>RequestDispatcher</code> obtained via 
    * <code>getRequestDispatcher()</code>, the <code>ServletRequest</code> 
    * object has its path elements and parameters adjusted to match
    * the path of the target resource.
    *
    * <p><code>forward</code> should be called before the response has been 
    * committed to the client (before response body output has been flushed).  
    * If the response already has been committed, this method throws
    * an <code>IllegalStateException</code>.
    * Uncommitted output in the response buffer is automatically cleared 
    * before the forward.
    *
    * <p>The request and response parameters must be the same
    * objects as were passed to the calling servlet's service method.
    *
    *
    * @param request		a {@link Pi3HttpServletRequest} object
    *				that represents the request the client
    * 				makes of the servlet
    *
    * @param response		a {@link Pi3HttpServletResponse} object
    *				that represents the response the servlet
    *				returns to the client
    *
    * @exception ServletException	if the target resource throws this exception
    *
    * @exception IOException	if the target resource throws this exception
    *
    * @exception IllegalStateException	if the response was already committed
    *
    */

    public void forward(ServletRequest request, ServletResponse response)
		throws ServletException, IOException {

        if ((req != null) && !request.equals(req)) {
		    ServletException e = new ServletException("Illegal ServletRequest used");
			throw e;
		}
		
		if (((Pi3HttpServletResponse)response).isCommitted()) {
		    IllegalStateException e = new IllegalStateException();
			throw e;
		}

        Servlet servlet = getServlet(url, request);
		if (servlet != null) {
		    try {
			    if (servlet instanceof SingleThreadModel)
					synchronized(servlet) {				
						servlet.service(request, response);
					}
				else 
					servlet.service(request, response);
				
				StringBuffer sb = new StringBuffer(servlet.getClass().getName());
				sb.append(".service() method invoked.");
				servlet.getServletConfig().getServletContext().log(sb.toString());
		    } catch(ServletException e) {
                throw e;
            }
		} else {
		    doForward(url, request, response);
			response.getOutputStream().close();
		};
	}




    /**
     *
     * Includes the content of a resource (servlet, JSP page,
     * HTML file) in the response. In essence, this method enables 
     * programmatic server-side includes.
     *
     * <p>The {@link Pi3HttpServletResponse} object has its path elements
     * and parameters remain unchanged from the caller's. The included
     * servlet cannot change the response status code or set headers;
     * any attempt to make a change is ignored.
     *
     * <p>The request and response parameters must be the same
     * objects as were passed to the calling servlet's service method.
     *
     *
     * @param request 			a {@link Pi3HttpServletRequest} object 
     *					that contains the client's request
     *
     * @param response 			a {@link Pi3HttpServletResponse} object 
     * 					that contains the servlet's response
     *
     * @exception ServletException 	if the included resource throws this exception
     *
     * @exception IOException 		if the included resource throws this exception
     *
     *
     */
     
    public void include(ServletRequest request, ServletResponse response)
	throws ServletException, IOException {

        if ((req != null) && !request.equals(req)) {
		    ServletException e = new ServletException("Illegal ServletRequest used");
			throw e;
		};

        Servlet servlet = getServlet(url, request);
		if (servlet != null) {
		    try {
			    if (servlet instanceof SingleThreadModel)
					synchronized(servlet) {				
						servlet.service(request, response);
					}
				 else
				 	servlet.service(request, response);

				StringBuffer sb = new StringBuffer(servlet.getClass().getName());
				sb.append(".service() method invoked.");
				servlet.getServletConfig().getServletContext().log(sb.toString());			
		    } catch(ServletException e) {
                throw e;
            }
		} else {
		    doForward(url, request, response);
		};
    }


    static {
        System.loadLibrary("Pi3ServletJNI");
    }
}
